import logging
import subprocess
from pathlib import Path

import uproot
from pandas import DataFrame
from ROOT import TFile  # type: ignore

from flux_tool.config import AnalysisConfig
from flux_tool.flux_systematics_analysis import FluxSystematicsAnalysis


class Exporter:
    __slots__ = ("nominal_samples", "products", "products_file")

    def __init__(self, cfg: AnalysisConfig, ana: FluxSystematicsAnalysis) -> None:
        self.nominal_samples = {
            "fhc": cfg.samples["fhc"]["nominal"],
            "rhc": cfg.samples["rhc"]["nominal"],
        }
        self.products_file = Path(cfg.products_file)
        self.products = ana.get_products()
        self.init_products_file()

    def init_products_file(self) -> None:
        logging.info(f"Creating output file:\n  {self.products_file}")

        self.products_file.unlink(missing_ok=True)

        if self.nominal_samples["fhc"] is not None:
            cmd1 = f"rootmkdir -p {self.products_file}:ppfx_output/fhc".split()
            subprocess.run(cmd1)
        if self.nominal_samples["rhc"] is not None:
            cmd2 = f"rootmkdir -p {self.products_file}:ppfx_output/rhc".split()
            subprocess.run(cmd2)

    def export_ppfx_output(self) -> None:
        fhc_file = self.nominal_samples["fhc"]
        rhc_file = self.nominal_samples["rhc"]
        if fhc_file is None and rhc_file is None:
            print("Please provide at least one of FHC/RHC file path!")
            return
        if fhc_file is not None:
            logging.info(
                f"Copying PPFX output from {fhc_file.name} to {self.products_file.name}"
            )
            cmd1 = f"rootcp -r {fhc_file} {self.products_file}:ppfx_output/fhc/".split()
            subprocess.run(cmd1)
        if rhc_file is not None:
            logging.info(
                f"Copying PPFX output from {rhc_file.name} to {self.products_file.name}"
            )
            cmd2 = f"rootcp -r {rhc_file} {self.products_file}:ppfx_output/rhc/".split()
            subprocess.run(cmd2)

    def export_product(self, product, key: str) -> None:
        with uproot.update(self.products_file) as products:
            products[key] = product

    def export_products(self) -> None:
        logging.info(f"Writing analysis products to {self.products_file.name}")
        product_file = TFile(str(self.products_file), "update")
        for key, product in self.products.items():
            logging.debug(f"\t{key}")
            if isinstance(product, DataFrame):
                continue
            dirs = key.split("/")

            if len(dirs) == 1:
                product_file.WriteObject(product, key)
            else:
                subdirs = "/".join(dirs[:-1])

                if not product_file.Get(subdirs):
                    product_file.mkdir(subdirs)
                d = product_file.Get(subdirs)

                d.WriteObject(product, dirs[-1])

        logging.info("Export complete. Closing file...")
        product_file.Close()



import itertools
from pathlib import Path
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from flux_tool.vis_scripts.helper import save_figure
from flux_tool.vis_scripts.spectra_reader import SpectraReader
from flux_tool.vis_scripts.style import neutrino_labels


def plot_matrices(
    matrices: dict[str, Any],
    horn_currents: list[str],
    nbins: dict[str, int],
    vlim: tuple[float, float] | str = (-1, 1),
    which: Optional[str] = None,
):
    bin_ordering = [nbins["nue"], nbins["nuebar"], nbins["numu"], nbins["numubar"]]

    if len(horn_currents) == 2:
        bin_ordering += bin_ordering[:-1]

    line_positions = list(itertools.accumulate(bin_ordering))

    nue = neutrino_labels["nue"]
    nueb = neutrino_labels["nuebar"]
    numu = neutrino_labels["numu"]
    numub = neutrino_labels["numubar"]

    for key, mat in matrices.items():
        if which is not None and which not in key:
            continue

        fig, ax = plt.subplots()  # layout="constrained")

        ax.set_box_aspect(1)

        m = mat.to_numpy()[0]

        if isinstance(vlim, tuple):
            vmin, vmax = vlim
        elif vlim == "auto":
            vmax = np.amax(m)
            vmin = -vmax
        else:
            raise ValueError(f"Unrecognized argument passed to plot_matrices: {vlim=}")

        heatmap_kwargs = {
            "ax": ax,
            "cmap": "bwr",
            "cbar": True,
            "vmin": vmin,
            "vmax": vmax,
            "square": False,
            "cbar_kws": {"shrink": 0.81},
        }

        sns.heatmap(m, **heatmap_kwargs)

        ax.invert_yaxis()  # type: ignore
        # ax.set_aspect(m.shape[1] / m.shape[0])  # type: ignore

        ax.tick_params(  # type: ignore
            bottom=False,
            top=False,
            left=False,
            right=False,
            labelbottom=False,
            labelleft=False,
            which="both",
        )

        kwargs = {
            "transform": ax.transAxes,  # type: ignore
            "fontsize": 32,
            "fontweight": "bold",
            # "ha": "center",
            # "va": "top",
        }
        kwargs2 = {
            "transform": ax.transAxes,  # type: ignore
            "fontsize": 32,
            "fontweight": "bold",
            # "va": "center",
            # "ha": "right",
        }

        # ax.annotate(  # type: ignore
        #     "ICARUS Preliminary",
        #     (0.0, 1.0),
        #     xytext=(0, 3),
        #     xycoords="axes fraction",
        #     textcoords="offset points",
        #     ha="left",
        #     va="bottom",
        #     fontweight="bold",
        # )  # type: ignore

        if len(horn_currents) == 1:
            xpos = -0.015
            nu_coords = [[xpos, 0.09], [xpos, 0.25], [xpos, 0.50], [xpos, 0.85]]
            horn = horn_currents[0].upper()
            ax.text(0.50, -0.085, horn, ha="center", va="top", **kwargs)  # type: ignore
            ax.text(-0.085, 0.50, horn, rotation=90, ha="right", va="center", **kwargs2)  # type: ignore

            for (x, y), nu in zip(nu_coords, neutrino_labels.values()):
                ax.text(x, y, nu, ha="right", va="center", **kwargs)  # type: ignore
                ax.text(y, x, nu, ha="center", va="top", **kwargs)  # type: ignore

            for pos in line_positions:
                lw = 1
                ax.axvline(pos, color="k", lw=lw)
                ax.axhline(pos, color="k", lw=lw)
        else:
            ax.text(0.245, -0.085, "FHC", ha="center", va="top", **kwargs)  # type: ignore
            ax.text(0.765, -0.085, "RHC", ha="center", va="top", **kwargs)  # type: ignore
            ax.text(  # type: ignore
                -0.085,
                0.245,
                "FHC",
                rotation=90,
                ha="right",
                va="center",
                **kwargs2,  # type: ignore
            )
            ax.text(  # type: ignore
                -0.085,
                0.765,
                "RHC",
                rotation=90,
                ha="right",
                va="center",
                **kwargs2,  # type: ignore
            )

            xpos = -0.015
            nu_coords = [
                [xpos, 0.05, nue],
                [xpos, 0.18, nueb],
                [xpos, 0.31, numu],
                [xpos, 0.44, numub],
                [xpos, 0.57, nue],
                [xpos, 0.70, nueb],
                [xpos, 0.83, numu],
                [xpos, 0.96, numub],
            ]

            for x, y, nu in nu_coords:
                ax.text(x, y, nu, ha="right", va="center", **kwargs2)  # type: ignore
                ax.text(y, x, nu, ha="center", va="top", **kwargs)  # type: ignore

            for i, pos in enumerate(line_positions):
                lw = 2 if i == 3 else 1
                ax.axvline(pos, color="k", lw=lw)
                ax.axhline(pos, color="k", lw=lw)

        yield key, fig


def plot_hadron_correlation_matrices(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    which: Optional[str] = None,
):
    horn_currents = reader.horn_current
    matrices = reader.hadron_correlation_matrices

    nbins = {k: len(v) - 1 for k, v in reader.binning.items()}

    figures = plot_matrices(matrices, horn_currents, nbins, which=which)

    if output_dir is not None:
        for key, fig in figures:
            category = key.split("/")[1]
            file_stem = f"{category}_correlation_matrix"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)  # type: ignore
            plt.close(fig)


def plot_hadron_covariance_matrices(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    which: Optional[str] = None,
):
    horn_currents = reader.horn_current
    matrices = reader.hadron_covariance_matrices

    nbins = {k: len(v) - 1 for k, v in reader.binning.items()}

    figures = plot_matrices(matrices, horn_currents, nbins, vlim="auto", which=which)

    if output_dir is not None:
        for key, fig in figures:
            category = key.split("/")[1]
            file_stem = f"{category}_covariance_matrix"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)  # type: ignore
            plt.close(fig)


def plot_beam_correlation_matrices(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    which: Optional[str] = None,
):
    horn_currents = reader.horn_current
    matrices = reader.beam_correlation_matrices

    nbins = {k: len(v) - 1 for k, v in reader.binning.items()}

    figures = plot_matrices(matrices, horn_currents, nbins, which=which)

    if output_dir is not None:
        for key, fig in figures:
            category = key
            if "/" in category:
                category = category.split("/")[1]
            category = category.split("_", 1)[1]
            file_stem = f"{category}_correlation_matrix"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)  # type: ignore
            plt.close(fig)


def plot_beam_covariance_matrices(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    which: Optional[str] = None,
):
    horn_currents = reader.horn_current
    matrices = reader.beam_covariance_matrices

    nbins = {k: len(v) - 1 for k, v in reader.binning.items()}

    figures = plot_matrices(matrices, horn_currents, nbins, vlim="auto", which=which)

    if output_dir is not None:
        for key, fig in figures:
            category = key
            if "/" in category:
                category = category.split("/")[1]
            category = category.split("_", 1)[1]
            file_stem = f"{category}_covariance_matrix"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)  # type: ignore
            plt.close(fig)



from dataclasses import dataclass, field
from functools import cached_property
from uuid import uuid4

import numpy as np
import pandas as pd
from ROOT import TF1, TH1D

from flux_tool.config import AnalysisConfig  # type: ignore


@dataclass(repr=False)
class FluxUniverseFit:
    flux_universes: pd.Series
    _flux_th1: TH1D = field(init=False)
    _fit_function: TF1 = field(init=False)

    def __post_init__(self) -> None:
        self._fit_function, self._flux_th1 = self.fit()

    def fit(self) -> tuple[TF1, TH1D]:
        hist_title = ";#phi_{#nu} (m^{-2} POT^{-1});PPFX Universes"
        n_universes = self.flux_universes.shape[0]
        nbins = int(np.ceil(np.sqrt(n_universes))) + 2
        xmin = np.amin(self.flux_universes)
        xmax = np.amax(self.flux_universes)

        flux_th1 = TH1D(str(uuid4()), hist_title, nbins, xmin, xmax)

        for uni in self.flux_universes:
            flux_th1.Fill(uni)

        fit_function = TF1(str(uuid4()), "gaus", xmin, xmax)

        flux_th1.Fit(fit_function, "0Q")

        return fit_function, flux_th1

    def eval_fit_function(self, x) -> float:
        return self._fit_function.Eval(x)

    @cached_property
    def flux_histogram(self) -> TH1D:
        return self._flux_th1

    @cached_property
    def universe_mean(self) -> float:
        # return np.mean(self.flux_universes)
        return self.flux_universes.mean()  # type: ignore

    @cached_property
    def universe_sigma(self) -> float:
        # return np.std(self.flux_universes)
        return self.flux_universes.std()  # type: ignore

    @cached_property
    def fit_mean(self) -> float:
        return self._fit_function.GetParameter(1)

    @cached_property
    def fit_sigma(self) -> float:
        return self._fit_function.GetParameter(2)

    @cached_property
    def mean_fractional_error(self) -> float:
        frac_err = np.abs(self.fit_mean - self.universe_mean) / self.universe_mean
        return frac_err

    @cached_property
    def sigma_fractional_error(self) -> float:
        frac_err = np.abs(self.fit_sigma - self.universe_sigma) / self.universe_sigma
        return frac_err

    @cached_property
    def chi2ndf(self) -> float:
        return self._fit_function.GetChisquare() / self._fit_function.GetNDF()


@dataclass(repr=False)
class HadronProductionSystematics:
    ppfx_dataframe: pd.DataFrame
    nominal_dataframe: pd.DataFrame
    cfg: AnalysisConfig
    _flux_pt: pd.DataFrame = field(init=False)
    _nom_pt: pd.DataFrame = field(init=False)

    def __post_init__(self) -> None:
        index = ("category", "horn_polarity", "neutrino_mode", "bin", "universe")
        flux_pt = pd.pivot_table(self.ppfx_dataframe, index=index, values="flux")
        nom_pt = pd.pivot_table(
            self.nominal_dataframe.loc[
                (self.nominal_dataframe["run_id"] == "nominal")
                & (self.nominal_dataframe["category"] == "nominal")
            ],
            values="flux",
            index=("horn_polarity", "neutrino_mode", "bin"),
        )
        self._flux_pt = flux_pt["flux"]
        self._nom_pt = nom_pt["flux"]

    @cached_property
    def ppfx_corrected_flux(self) -> pd.DataFrame:
        mean = self._flux_pt.groupby(level=self._flux_pt.index.names[:4]).mean()
        sigma = self._flux_pt.groupby(level=self._flux_pt.index.names[:4]).std()

        return pd.concat([mean, sigma], axis=1, keys=["mean", "sigma"])  # type: ignore

    @cached_property
    def ppfx_flux_weights(self) -> pd.DataFrame:
        total_correction = self.ppfx_corrected_flux.loc["total", "mean"]
        weights = total_correction / self._nom_pt
        return weights.fillna(0)

    @cached_property
    def covariance_matrices(self) -> pd.DataFrame:
        mean = self.ppfx_corrected_flux["mean"]

        pt_frac = self._flux_pt.div(mean)

        group_abs = self._flux_pt.unstack(
            ("horn_polarity", "neutrino_mode", "bin")
        ).groupby(level="category")

        group_frac = pt_frac.unstack(("horn_polarity", "neutrino_mode", "bin")).groupby(
            level="category"
        )

        cov_abs = group_abs.cov()
        cov_frac = group_frac.cov()

        cov = pd.concat(
            [cov_frac, cov_abs],
            keys=["fractional", "absolute"],
            names=["scale"] + cov_abs.index.names,
        ).fillna(0)

        return cov

    @cached_property
    def correlation_matrices(self) -> pd.DataFrame:
        group = self._flux_pt.unstack(
            ("horn_polarity", "neutrino_mode", "bin")
        ).groupby(level="category")

        corr = group.corr().fillna(0)

        return corr

    @cached_property
    def fractional_uncertainties(self) -> pd.DataFrame:
        cov = self.covariance_matrices.loc["fractional"]

        uncerts = (
            cov.groupby("category")
            .apply(lambda x:  pd.Series(np.sqrt(np.diag(x)), index=cov.columns))
            # TODO remove future_stack=True in pandas ^3.0.0
            .stack(["horn_polarity", "neutrino_mode", "bin"], future_stack=True)
            .sort_index()
        )

        return uncerts

    @cached_property
    def flux_fit_results(self) -> dict[tuple[str, str, int], FluxUniverseFit]:
        bins_df = pd.pivot_table(
            data=self.ppfx_dataframe.query("category == 'total'"),
            index=["horn_polarity", "neutrino_mode", "bin", "universe"],
            values="flux",
        )["flux"]

        flux_fits = {}
        for idx, b in bins_df.groupby(level=["horn_polarity", "neutrino_mode", "bin"]):
            flux_fits |= {idx: FluxUniverseFit(b.droplevel(0))}

        return flux_fits



from functools import partial
from dataclasses import dataclass, field
from itertools import product
from pathlib import Path
from typing import Any, Callable, Iterator, Optional

import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np
from matplotlib.figure import Figure
from numpy.typing import NDArray
from ROOT import TH1D  # type: ignore

from flux_tool.vis_scripts.helper import absolute_uncertainty, save_figure
from flux_tool.vis_scripts.spectra_reader import SpectraReader
from flux_tool.vis_scripts.style import (beam_syst_colors, beam_syst_labels,
                                         beam_syst_lines, neutrino_labels,
                                         ppfx_colors, ppfx_labels, ppfx_lines,
                                         ppfx_mesinc_colors, xlabel_enu)


def _default_legend_kwargs() -> dict[str, Any]:
    return {
        "loc": "upper center",
        "ncol": 2,
        "fontsize": 20,
        "columnspacing": 0.8,
    }

@dataclass(slots=True, frozen=True)
class PlotComponents:
    horn: str
    nu: str
    total_uncertainty: TH1D
    output_name: str
    xlim: tuple[float, float] = field(default_factory=tuple)
    ylim: tuple[float, float] = field(default_factory=tuple)
    uncertainties: list[TH1D] = field(default_factory=list)
    labels: list[str] = field(default_factory=list)
    colors: list[str] = field(default_factory=list)
    linestyles: list[str] = field(default_factory=list)
    legend_kwargs: dict[str, Any] = field(default_factory=_default_legend_kwargs)

def create_figure(comps: PlotComponents, label_drawer: Optional[partial] = None) -> Figure | None:
    n_spectra = len(comps.uncertainties)

    if n_spectra == 0:
        return

    # half = n_spectra // 2

    # ls = (half) * ["-"] + (n_spectra - half) * ["--"] if n_spectra > 6 else "-"

    fig, ax = plt.subplots(figsize=[10,10])  # , layout="constrained")

    ax.set_box_aspect(1)

    if comps.total_uncertainty is not None:
        hep.histplot(
            ax=ax,
            H=comps.total_uncertainty,
            yerr=False,
            histtype="step",
            linestyle="-",
            lw=3,
            color="k",
            label="Total",
            edges=False,
        )
    colors = comps.colors if len(comps.colors) > 0 else None
    linestyles = comps.linestyles if len(comps.linestyles) > 0 else None
    hep.histplot(
        ax=ax,
        H=comps.uncertainties,
        yerr=False,
        histtype="step",
        edges=False,
        color=colors,
        lw=3,
        ls=linestyles,
        label=comps.labels,
    )

    header = {
        "fhc": f"FHC {neutrino_labels[comps.nu]}",
        "rhc": f"RHC {neutrino_labels[comps.nu]}",
    }

    # ax.set_yticks(np.arange(0, 0.30, step=0.02))
    ax.set_xlim(*comps.xlim)
    ax.set_ylim(*comps.ylim)
    ax.legend(**comps.legend_kwargs)
    ax.set_xlabel(xlabel_enu)
    ax.set_ylabel(r"Fractional Uncertainty $\mathrm{\left( \sigma / \phi \right)}$")

    if label_drawer is not None:
        label_drawer(ax=ax, rlabel=header[comps.horn])

    return fig


def plot_uncertainties(
    reader: SpectraReader,
    fn: Callable[[SpectraReader, tuple[float, float], tuple[float, float]], Iterator],
    output_dir: Optional[Path] = None,
    xlim: tuple[float, float] = (0, 20),
    ylim: tuple[float, float] = (0, 0.27),
    label_drawer: Optional[partial] = None,
    flux_overlay: bool = False,
):
    for comps in fn(reader, xlim, ylim):
        fig = create_figure(comps, label_drawer)
        if fig is None:
            continue

        if flux_overlay:
            create_flux_overlay(reader, comps.horn, comps.nu)

        if output_dir is not None:
            fig_name = comps.output_name
            tex_label = fig_name
            tex_caption = ""

            save_figure(fig, fig_name, output_dir, tex_caption, tex_label)

        plt.close(fig)


def rebin_within_xlim(hist: TH1D, binning: NDArray, xlim: tuple[float, float]) -> TH1D:
    new_binning = binning[(binning >= xlim[0]) & (binning <= xlim[1])]
    hist = hist.Rebin(len(new_binning) - 1, hist.GetName(), new_binning)
    return hist


def plot_hadron_fractional_uncertainties(
    reader: SpectraReader,
    xlim: tuple[float, float],
    ylim: tuple[float, float],
) -> Iterator[PlotComponents]:
    ppfx_correction = reader.ppfx_correction
    hadron_uncertainties = reader.hadron_uncertainties

    for horn, nu in reader.horns_and_nus:
        flux = ppfx_correction[f"htotal_{horn}_{nu}"].to_pyroot()
        flux = rebin_within_xlim(flux, reader.binning[nu], xlim)

        total_uncertainty = hadron_uncertainties[
            f"total/hfrac_hadron_total_{horn}_{nu}"
        ].to_pyroot()

        total_uncertainty = rebin_within_xlim(
            total_uncertainty, reader.binning[nu], xlim
        )

        uncerts = {
            k.split("/")[0]: v.to_pyroot()
            for k, v in hadron_uncertainties.items()
            if k.endswith(nu)
            and horn in k
            and "total" not in k
            and "projectile" not in k
            and "daughter" not in k
            and "incoming" not in k
            and "outgoing" not in k
            and "inC" not in k
            and "outC" not in k
            and "nua_other" not in k
        }

        for k, h in uncerts.items():
            uncerts[k] = rebin_within_xlim(h, reader.binning[nu], xlim)

        sorted_uncerts = sort_uncertainties(uncerts, flux)

        hadron_labels = [ppfx_labels[k] for k in sorted_uncerts]

        colors = [ppfx_colors[k] for k in sorted_uncerts]

        ls = [ppfx_lines[k] for k in sorted_uncerts]

        yield PlotComponents(
            horn=horn,
            nu=nu,
            total_uncertainty=total_uncertainty,
            output_name=f"{horn}_{nu}_hadron_fractional_uncertainties",
            xlim=xlim,
            ylim=ylim,
            uncertainties=list(sorted_uncerts.values()),
            labels=hadron_labels,
            colors=colors,
            linestyles=ls,
        )


def plot_hadron_fractional_uncertainties_mesinc_breakout(
    reader: SpectraReader,
    xlim: tuple[float, float],
    ylim: tuple[float, float],
) -> Iterator[PlotComponents]:
    # version = {"projectile": "incoming", "daughter": "outgoing"}

    all_versions = product(
        reader.horns_and_nus,
        # ["daughter", "projectile"],
        ["incoming", "outgoing"],
    )
    all_versions_list = [(*v[0], v[1]) for v in all_versions]

    ppfx_correction = reader.ppfx_correction
    hadron_uncertainties = reader.hadron_uncertainties

    for horn, nu, ver in all_versions_list:
        flux = ppfx_correction[f"htotal_{horn}_{nu}"].to_pyroot()
        flux = rebin_within_xlim(flux, reader.binning[nu], xlim)

        # total_uncertainty = hadron_uncertainties[
        #     f"total/hfrac_hadron_total_{horn}_{nu}"
        # ].to_pyroot()

        # total_uncertainty = rebin_within_xlim(
        #     total_uncertainty, reader.binning[nu], xlim
        # )

        uncerts = {
            k.split("/")[0]: v.to_pyroot()
            for k, v in hadron_uncertainties.items()
            if k.endswith(nu)
            and horn in k
            and "total" not in k
            and ver in k
            and "mesinc/" not in k
        }

        for k, h in uncerts.items():
            uncerts[k] = rebin_within_xlim(h, reader.binning[nu], xlim)

        sorted_uncerts = sort_uncertainties(uncerts, flux)

        hadron_labels = [ppfx_labels[k] for k in sorted_uncerts]

        colors = [ppfx_mesinc_colors[k] for k in sorted_uncerts]

        yield PlotComponents(
            horn=horn,
            nu=nu,
            total_uncertainty=None,
            # f"{version[ver]}/{horn}_{nu}_{version[ver]}_hadron_fractional_uncertainties",
            output_name=f"{horn}_{nu}_{ver}_mesinc_fractional_uncertainties",
            xlim=xlim,
            ylim=( 0, 0.15 ),
            uncertainties=list(sorted_uncerts.values()),
            labels=hadron_labels,
            colors=colors,
        )


def plot_hadron_fractional_uncertainties_nua_breakout(
    reader: SpectraReader,
    xlim: tuple[float, float],
    ylim: tuple[float, float],
) -> Iterator[PlotComponents]:
    ppfx_correction = reader.ppfx_correction
    hadron_uncertainties = reader.hadron_uncertainties

    for horn, nu in reader.horns_and_nus:
        flux = ppfx_correction[f"htotal_{horn}_{nu}"].to_pyroot()
        flux = rebin_within_xlim(flux, reader.binning[nu], xlim)

        total_uncertainty = hadron_uncertainties[
            f"total/hfrac_hadron_total_{horn}_{nu}"
        ].to_pyroot()

        total_uncertainty = rebin_within_xlim(
            total_uncertainty, reader.binning[nu], xlim
        )

        # uncerts = {
        #         "nua": hadron_uncertainties[f"nua/hfrac_hadron_nua_{horn}_{nu}"].to_pyroot(),
        #         "nua_datavol": hadron_uncertainties[f"nua_datavol/hfrac_hadron_nua_datavol_{horn}_{nu}"].to_pyroot(),
        #         "nua_datavol_negxF": hadron_uncertainties[f"nua_datavol_negxF/hfrac_hadron_nua_datavol_negxF_{horn}_{nu}"].to_pyroot(),
        #         "nua_othervol": hadron_uncertainties[f"nua_othervol/hfrac_hadron_nua_othervol_{horn}_{nu}"].to_pyroot(),
        #         "nua_other": hadron_uncertainties[f"nua_other/hfrac_hadron_nua_other_{horn}_{nu}"].to_pyroot(),
        # }
        uncerts = {
                "nua": hadron_uncertainties[f"nua/hfrac_hadron_nua_{horn}_{nu}"].to_pyroot(),
                "nua_inC_inPS": hadron_uncertainties[f"nua_inC_inPS/hfrac_hadron_nua_inC_inPS_{horn}_{nu}"].to_pyroot(),
                "nua_inC_OOPS": hadron_uncertainties[f"nua_inC_OOPS/hfrac_hadron_nua_inC_OOPS_{horn}_{nu}"].to_pyroot(),
                "nua_outC_Ascale": hadron_uncertainties[f"nua_outC_Ascale/hfrac_hadron_nua_outC_Ascale_{horn}_{nu}"].to_pyroot(),
                "nua_outC_OOPS": hadron_uncertainties[f"nua_outC_OOPS/hfrac_hadron_nua_outC_OOPS_{horn}_{nu}"].to_pyroot(),
                "nua_other": hadron_uncertainties[f"nua_other/hfrac_hadron_nua_other_{horn}_{nu}"].to_pyroot(),
        }

        for k, h in uncerts.items():
            uncerts[k] = rebin_within_xlim(h, reader.binning[nu], xlim)

        sorted_uncerts = sort_uncertainties(uncerts, flux)

        hadron_labels = [ppfx_labels[k] for k in sorted_uncerts]

        colors = [ppfx_colors[k] if k != "nua" else "k" for k in sorted_uncerts]

        # lines = [ppfx_lines[k] for k in sorted_uncerts]

        yield PlotComponents(
            horn=horn,
            nu=nu,
            total_uncertainty=None,
            output_name=f"{horn}_{nu}_hadron_fractional_uncertainties_nua_only",
            xlim=xlim,
            ylim=ylim,
            uncertainties=list(sorted_uncerts.values()),
            labels=hadron_labels,
            colors=colors,
            # linestyles=lines
            legend_kwargs={
                "loc": "upper right",
                "ncol": 1,
                "fontsize": 22,
                "columnspacing": 0.8,
                "reverse": True,
                }
        )


def plot_beam_fractional_uncertainties(
    reader: SpectraReader,
    xlim: tuple[float, float],
    ylim: tuple[float, float],
) -> Iterator[PlotComponents]:
    ppfx_correction = reader.ppfx_correction
    uncertainties = reader.beam_uncertainties

    for horn, nu in reader.horns_and_nus:
        flux = ppfx_correction[f"htotal_{horn}_{nu}"].to_pyroot()
        flux = rebin_within_xlim(flux, reader.binning[nu], xlim)

        total_uncertainty = uncertainties[
            f"total/hfrac_beam_total_{horn}_{nu}"
        ].to_pyroot()

        total_uncertainty = rebin_within_xlim(
            total_uncertainty, reader.binning[nu], xlim
        )

        uncerts = {
            k.split("/")[0]: v.to_pyroot()
            for k, v in uncertainties.items()
            if k.endswith(nu)
            and horn in k
            and "total" not in k
            and "beam_power" not in k
        }
        for k, h in uncerts.items():
            uncerts[k] = rebin_within_xlim(h, reader.binning[nu], xlim)

        sorted_uncerts = sort_uncertainties(uncerts, flux)

        labels = [beam_syst_labels[k] for k in sorted_uncerts]

        colors = [beam_syst_colors[k] for k in sorted_uncerts]

        ls = [beam_syst_lines[k] for k in sorted_uncerts]

        comps = PlotComponents(
            horn,
            nu,
            total_uncertainty,
            f"{horn}_{nu}_beam_fractional_uncertainties",
            xlim,
            ylim,
            list(sorted_uncerts.values()),
            labels,
            colors,
            ls,
        )

        comps.legend_kwargs["fontsize"] = 18

        yield comps


def create_flux_overlay(reader: SpectraReader, horn: str, nu: str) -> None:
    flux = reader.ppfx_correction[f"htotal_{horn}_{nu}"].to_pyroot()
    ax1 = plt.gca()
    ax2 = ax1.twinx()

    max = flux.GetMaximum()

    h = hep.histplot(
        flux,
        ax=ax2,
        histtype="fill",
        yerr=False,
        label=("Flux Shape (A.U.)"),
        ls="-.",
        color="gray",
        edgecolor="k",
        lw=3,
        zorder=0,
        alpha=0.3,
    )

    handles, labels = ax1.get_legend_handles_labels()

    handles += [h[0][0]]
    labels += ["Flux Shape (A.U.)"]

    ax1.legend(handles, labels, ncol=2, loc="upper center", fontsize=26)

    ax2.set_ylim(0, 1.4 * max)

    ax2.get_yaxis().set_visible(False)


def plot_beam_systematic_shifts(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    xlim: tuple[float, float] = (0, 20),
    # ylim: tuple[float, float] = (0, 0.1),
) -> None:
    shifts = reader.beam_systematic_shifts

    for key, shift in shifts.items():
        if "beam_power" in key:
            continue
        tmp, horn, nu = key.rsplit("_", 2)
        _, _, syst = tmp.split("_", 2)

        label = beam_syst_labels[syst]
        color = beam_syst_colors[syst]

        stats, bins = reader[f"statistical_uncertainties/hstat_{horn}_{nu}"].to_numpy()  # type: ignore

        fig, ax = plt.subplots(layout="constrained", figsize=(11, 11))

        hep.histplot(
            H=shift,
            label=label,
            ax=ax,
            yerr=False,
            edges=False,
            lw=3,
            zorder=10,
            color=color,
        )
        hep.histplot(
            H=stats,
            bins=bins,
            ax=ax,
            histtype="fill",
            label=r"$\mathrm{\sigma}_\mathsf{stat}$",
            color="C0",
            alpha=0.4,
            zorder=0,
        )
        hep.histplot(
            H=-1 * stats,
            bins=bins,
            ax=ax,
            histtype="fill",
            color="C0",
            alpha=0.4,
            zorder=0,
        )

        ax.axhline(0, ls="--", color="k", zorder=1)

        ax.set_xlim(xlim)
        ax.set_ylim(-0.1, 0.1)
        ax.legend(loc="best")
        ax.set_xlabel(xlabel_enu)
        ax.set_ylabel(
            r"$\mathrm{\phi}_x - \mathrm{\phi}_\mathsf{nom}$ / $\mathrm{\phi}_\mathsf{nom}$"
        )

        hep.label.exp_label(
            exp="", llabel="", rlabel=f"{horn.upper()} {neutrino_labels[nu]}"
        )

        if output_dir is not None:
            prefix = f"{horn}_{nu}"
            fig_name = f"{prefix}_{syst}_systematic_shift"
            tex_label = fig_name
            tex_caption = ""

            save_figure(fig, fig_name, output_dir, tex_caption, tex_label)

        plt.close(fig)


def sort_uncertainties(uncertainties: dict[str, TH1D], flux: TH1D) -> dict[str, TH1D]:
    absolute_uncertainties = {
        k: absolute_uncertainty(flux, h).Integral(1, 11)
        for k, h in uncertainties.items()
    }

    sorted_uncertainties = {
        k: v
        for k, v in sorted(
            uncertainties.items(),
            key=lambda kv: absolute_uncertainties[kv[0]],
            reverse=True,
        )
    }

    return sorted_uncertainties



import logging
from typing import Any, Callable, Optional

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from ROOT import TH2D, TAxis  # type: ignore

from flux_tool import uncertainty
from flux_tool.beam_focusing_systematics import BeamFocusingSystematics
from flux_tool.config import AnalysisConfig
from flux_tool.hadron_production_systematics import HadronProductionSystematics
from flux_tool.helpers import (calculate_correlation_matrix,
                               convert_groups_to_dict, convert_pandas_to_th1,
                               convert_pandas_to_th2,
                               convert_symmetric_ndarray_to_tmatrix)
from flux_tool.principal_component_analysis import PCA


class FluxSystematicsAnalysis:
    __slots__ = (
        "beam_systematics",
        "beam_systematics_is_initialized",
        "bin_edges",
        "cfg",
        "flux_prediction",
        "hadron_systematics",
        "horn_modes",
        "nominal_flux_df",
        "pca_eigen_values",
        "pca_components",
        "pca_covariance_matrix",
        "ppfx_correction_df",
        "statistical_uncertainties",
        "stat_uncert_matrix",
        "th2_bins",
        "total_covariance_matrix",
    )

    def __init__(
        self,
        nominal_flux_df: pd.DataFrame,
        ppfx_correction_df: pd.DataFrame,
        bin_edges: dict[str, np.ndarray],
        cfg: AnalysisConfig
    ) -> None:
        self.nominal_flux_df = nominal_flux_df
        self.ppfx_correction_df = ppfx_correction_df
        self.bin_edges = bin_edges
        self.cfg = cfg
        self.beam_systematics_is_initialized = (
            len(self.nominal_flux_df["run_id"].unique()) > 1
        )
        self.horn_modes = list(self.nominal_flux_df["horn_polarity"].unique())

    def run(self, pca_threshold: float = 1) -> None:
        logging.info("\n=============== BEGINNING ANALYSIS ===============")

        logging.info(
            "Calculating flux correction and hadron production systematic uncertainties..."
        )
        self.hadron_systematics = HadronProductionSystematics(
            self.ppfx_correction_df, self.nominal_flux_df, self.cfg
        )

        logging.info("Reading statistical uncertainties...")
        statistical_uncertainties = uncertainty.extract_statistical_uncertainties(
            self.nominal_flux_df, self.hadron_systematics.ppfx_flux_weights
        )

        statistical_uncertainties_fraction = (
            uncertainty.extract_statistical_uncertainties(
                self.nominal_flux_df,
                self.hadron_systematics.ppfx_flux_weights,
                normalized=True,
            )
        )

        diag = np.power(np.diag(statistical_uncertainties), 2)

        self.stat_uncert_matrix = pd.DataFrame(
            diag,
            index=statistical_uncertainties.index,
            columns=statistical_uncertainties.index,
        )

        self.statistical_uncertainties = pd.concat(
            [statistical_uncertainties, statistical_uncertainties_fraction],
            keys=["absolute", "fractional"],
            names=["scale"] + statistical_uncertainties.index.names,
        )

        self.th2_bins = np.arange(
            self.hadron_systematics.correlation_matrices.loc["total"].shape[0] + 1
        )

        if self.beam_systematics_is_initialized:
            logging.info("Calculating beamline focusing systematic uncertainties...")
            self.beam_systematics = BeamFocusingSystematics(
                beam_flux_df=self.nominal_flux_df,
                bin_edges=self.bin_edges,
                smoothing=True,
            )

        total_cov = self.hadron_systematics.covariance_matrices.loc[
            ("fractional", "total")
        ]

        pca = PCA(total_cov, threshold=pca_threshold)  # type: ignore

        logging.info("Performing PCA on hadron production covariance matrix...")
        pca.fit()

        self.pca_covariance_matrix = pd.DataFrame(
            pca.new_covariance_matrix, index=total_cov.index, columns=total_cov.columns
        )

        self.pca_eigen_values = pca.eigenvalues_df
        self.pca_components = pca.principal_component_df

        self.total_covariance_matrix = self._total_covariance_matrix

        self.flux_prediction = self._flux_prediction

    def rescale_matrix(self, matrix, fractional=True):
        """Helper function to convert between absolute and fractional scales of the covariance matrices.
        If fractional=True (default), then the input matrix is presumed to be in the fractional scale and will be multipled by flux to return to absolute scale.
        """
        flux = self.hadron_systematics.ppfx_corrected_flux.loc["total", "mean"]
        scale_factor = np.outer(flux, flux)
        if not fractional:
            scale_factor = np.reciprocal(
                scale_factor,
                out=np.zeros(shape=scale_factor.shape),
                where=scale_factor != 0,
            )
        return matrix * scale_factor

    @property
    def xaxis_variable_bins(self) -> TAxis:
        xbins = self.bin_edges["numu"]
        nbins = xbins.shape[0] - 1

        axis = TAxis(nbins, xbins)
        axis.SetTitle("E_{#nu} [GeV]")

        return axis

    @property
    def matrix_taxis(self) -> TAxis:
        matrix = self.total_covariance_matrix
        rows = matrix.index
        columns = matrix.columns

        xbins = np.arange(0, len(rows) + 1, dtype=float)
        nbinsx = xbins.shape[0] - 1

        taxis = TAxis(nbinsx, xbins)

        it = zip(enumerate(rows), enumerate(columns))

        for (ii, row), (jj, col) in it:
            labelx = f"{row[0]}-{row[1]}-{row[2]}"
            labely = f"{col[0]}-{col[1]}-{col[2]}"
            taxis.SetBinLabel(ii + 1, labelx)
            taxis.SetBinLabel(jj + 1, labely)

        return taxis

    @property
    def matrix_binning_str(self) -> str:
        index = self.total_covariance_matrix.index
        bins = self.bin_edges["numu"]

        nu_pdg = {"nue": 12, "nuebar": -12, "numu": 14, "numubar": -14}

        lines = ["variables: isRHC NeutrinoCode Enu Enu"]
        for horn, nu, b in index:
            is_RHC = 1 if horn == "rhc" else 0
            lines.append(f"{is_RHC} {nu_pdg[nu]} {bins[b-1]} {bins[b]}")

        return "\n".join(lines)

    @property
    def _total_covariance_matrix(self) -> pd.DataFrame:
        logging.info("Computing total covariance matrix...")
        hp_mat = self.hadron_systematics.covariance_matrices.loc["absolute", "total"]
        total_mat = hp_mat + self.stat_uncert_matrix
        if self.beam_systematics_is_initialized:
            beam_tot_cov = self.beam_systematics.total_covariance_matrix.loc["absolute"]
            beam_power_cov = self.beam_systematics.covariance_matrices.loc[
                "absolute", "beam_power"
            ]
            return total_mat + beam_tot_cov + beam_power_cov
        return total_mat

    @property
    def total_correlation_matrix(self) -> pd.DataFrame | NDArray[Any]:
        corr_mat = calculate_correlation_matrix(self.total_covariance_matrix)
        return corr_mat

    @property
    def _flux_prediction(self) -> pd.DataFrame:
        total_sigma = pd.Series(
            np.sqrt(np.diag(self.total_covariance_matrix)),
            index=self.total_covariance_matrix.index,
            name="sigma",
        )

        ppfx_mean = self.hadron_systematics.ppfx_corrected_flux.loc["total", "mean"]

        return pd.concat([ppfx_mean, total_sigma], axis=1)

    def total_uncertainties_in_range(
        self, elow, ehigh, mat: Optional[pd.DataFrame] = None
    ) -> pd.DataFrame:
        if mat is None:
            mat = self.total_covariance_matrix

        bin_slices = {}
        for nu, bins in self.bin_edges.items():
            bins_selected = np.where((bins >= elow) & (bins <= ehigh))[0]
            if bins_selected.shape == bins.shape:
                bin_slice = slice(None)
            else:
                bin_slice = slice(bins_selected[0], bins_selected[-1])
            bin_slices[nu] = bin_slice

        df_list = []

        for horn in self.horn_modes:
            index_slice = pd.IndexSlice[horn, :, bin_slices["numu"]]

            cov_mat = (
                mat.loc[index_slice, index_slice]
                .droplevel(level="horn_polarity", axis=0)
                .droplevel(level="horn_polarity", axis=1)
            )

            nue_flux = self.flux_prediction.loc[
                (horn, "nue", bin_slices["nue"]), "mean"
            ].sum()

            nuebar_flux = self.flux_prediction.loc[
                (horn, "nuebar", bin_slices["nuebar"]), "mean"
            ].sum()

            numu_flux = self.flux_prediction.loc[
                (horn, "numu", bin_slices["numu"]), "mean"
            ].sum()

            numubar_flux = self.flux_prediction.loc[
                (horn, "numubar", bin_slices["numubar"]), "mean"
            ].sum()

            total_nue_flux = nue_flux + nuebar_flux
            total_numu_flux = numu_flux + numubar_flux

            ratio_uncert = uncertainty.ratio_uncertainty(
                cov=cov_mat,
                total_nue_flux=total_nue_flux,
                total_numu_flux=total_numu_flux,
            )

            nue_uncert = uncertainty.flux_uncertainty(
                cov=cov_mat.loc["nue", "nue"], total_flux=nue_flux
            )

            nuebar_uncert = uncertainty.flux_uncertainty(
                cov=cov_mat.loc["nuebar", "nuebar"], total_flux=nuebar_flux
            )

            nue_nuebar_uncert = uncertainty.flux_uncertainty(
                cov=cov_mat.loc[slice("nue", "nuebar"), slice("nue", "nuebar")],
                total_flux=total_nue_flux,
            )

            numu_uncert = uncertainty.flux_uncertainty(
                cov=cov_mat.loc["numu", "numu"], total_flux=numu_flux
            )

            numubar_uncert = uncertainty.flux_uncertainty(
                cov=cov_mat.loc["numubar", "numubar"], total_flux=numubar_flux
            )

            numu_numubar_uncert = uncertainty.flux_uncertainty(
                cov=cov_mat.loc[slice("numu", "numubar"), slice("numu", "numubar")],
                total_flux=total_numu_flux,
            )

            res_dict = {
                "nue": [nue_uncert],
                "nuebar": [nuebar_uncert],
                "nue+nuebar": [nue_nuebar_uncert],
                "numu": [numu_uncert],
                "numubar": [numubar_uncert],
                "numu+numubar": [numu_numubar_uncert],
                "nue+nuebar/numu+numubar": [ratio_uncert],
            }

            df = pd.DataFrame(res_dict)
            df_list.append(df)

        return pd.concat(df_list, keys=["fhc", "rhc"]).droplevel(level=-1)

    @property
    def total_uncertainty_table(self) -> pd.DataFrame:
        enu_min, enu_max = 0, 20

        uncertainties = [
            self.total_uncertainties_in_range(
                enu_min,
                enu_max,
                self.hadron_systematics.covariance_matrices.loc["absolute", "total"],
            )
        ]

        if self.beam_systematics_is_initialized:
            uncertainties += [
                self.total_uncertainties_in_range(
                    enu_min,
                    enu_max,
                    self.beam_systematics.total_covariance_matrix.loc["absolute"],
                ),
                self.total_uncertainties_in_range(
                    enu_min,
                    enu_max,
                    self.beam_systematics.covariance_matrices.loc[
                        "absolute", "beam_power"
                    ],
                ),
            ]

        uncertainties += [
            self.total_uncertainties_in_range(
                enu_min, enu_max, self.stat_uncert_matrix
            ),
            self.total_uncertainties_in_range(enu_min, enu_max),
        ]

        table_keys = [
            "Hadron",
            "Beamline",
            "Beam Power Upgrade",
            "Statistical",
            "Total",
        ]

        uncert_table = pd.concat(uncertainties, keys=table_keys)

        uncert_table.columns = [
            r"$\nu_e$",
            r"$\bar{\nu}_e$",
            r"$\nu_e + \bar{\nu}_e$",
            r"$\nu_\mu$",
            r"$\bar{\nu}_\mu$",
            r"$\nu_\mu + \bar{\nu}_\mu$",
            r"$\frac{\nu_e + \bar{\nu}_e}{\nu_\mu + \bar{\nu}_\mu}$",
        ]

        return uncert_table

    @property
    def total_uncertainty_table_latex(self) -> str:
        table = self.total_uncertainty_table

        table *= 100  # convert to percentage
        table_str = table.to_latex(float_format="%.2f")
        table_str = table_str.replace("fhc", "FHC")
        table_str = table_str.replace("rhc", "RHC")

        return table_str

    @staticmethod
    def _export_matrices(
        matrix: pd.DataFrame,
        title_gen: str | Callable[[str], str],
    ) -> dict[str, TH2D]:
        if isinstance(title_gen, str):
            return {
                title_gen: convert_pandas_to_th2(matrix, hist_name=title_gen),
                title_gen.replace("hcov", "cov").replace(
                    "hcorr", "corr"
                ): convert_symmetric_ndarray_to_tmatrix(matrix.values),
            }

        export_dict = {}

        for category, hist in matrix.groupby(level=0):
            hist_title = title_gen(str(category))
            h = hist.droplevel(0)
            export_dict[hist_title] = convert_pandas_to_th2(h, hist_name=hist_title)
            mat_title = hist_title.replace("hcov", "cov").replace("hcorr", "corr")
            export_dict[mat_title] = convert_symmetric_ndarray_to_tmatrix(hist.values)

        return export_dict

    def _export_nominal_flux_df(self):
        pt = pd.pivot_table(
            self.nominal_flux_df,
            index=("run_id", "category", "horn_polarity", "neutrino_mode", "bin"),
            values=["flux", "stat_uncert"],
        )

        export_dict = {}

        for idx, spectra in pt.groupby(
            level=("run_id", "category", "horn_polarity", "neutrino_mode")
        ):
            run, cat, horn, nu = idx  # type: ignore

            directory = f"beam_samples/run_{run}/"

            hist_name = "h"

            if cat == "central_value":
                hist_name += "cv"
            elif cat == "nominal":
                hist_name += "nom"
            else:
                hist_name += f"nom_{cat}"
                directory += "parent/"

            hist_name += f"_{horn}_{nu}"

            th1 = convert_pandas_to_th1(
                series=spectra["flux"],
                bin_edges=self.bin_edges[nu],
                hist_name=directory + hist_name,
                uncerts=spectra["stat_uncert"],
            )
            th1.SetTitle(";E_{#nu} [GeV]; #Phi_{#nu} [m^{-2} POT^{-1}]")

            export_dict[directory + hist_name] = th1

        return export_dict

    def _export_flux_universes(self):
        df = self.ppfx_correction_df.drop("run_id", axis=1)
        df["bin"] = df["bin"].astype(str)
        df = pd.pivot_table(
            df,
            index="universe",
            columns=("category", "horn_polarity", "neutrino_mode", "bin"),
            values=["flux"],
        )["flux"]
        df.columns = df.columns.map("_".join)

        return {"flux_universes": df}

    def get_products(self) -> dict[str, Any]:
        product_dict = {}

        product_dict["matrix_axis"] = self.matrix_taxis

        product_dict["xaxis_variable_bins"] = self.xaxis_variable_bins

        stat_mat_title = "hstatistical_uncertainty_matrix"

        stat_uncert_mat = convert_pandas_to_th2(
            self.stat_uncert_matrix, hist_name=stat_mat_title
        )

        stat_uncert_tmatrix = convert_symmetric_ndarray_to_tmatrix(
            self.stat_uncert_matrix.values
        )

        product_dict[f"statistical_uncertainties/{stat_mat_title}"] = stat_uncert_mat

        product_dict[
            "statistical_uncertainties/statistical_uncertainty_matrix"
        ] = stat_uncert_tmatrix

        product_dict |= self._export_nominal_flux_df()

        product_dict |= self._export_flux_universes()

        levels = ["horn_polarity", "neutrino_mode"]
        three_levels = ["category"] + levels

        stat_uncrt_abs_groups = self.statistical_uncertainties.loc["absolute"].groupby(
            level=levels
        )

        stat_uncrt_frac_groups = self.statistical_uncertainties.loc[
            "fractional"
        ].groupby(level=levels)

        had_uncrt_groups = self.hadron_systematics.fractional_uncertainties.groupby(
            level=three_levels
        )

        flux_weights_groups = self.hadron_systematics.ppfx_flux_weights.groupby(
            level=levels
        )

        ppfx_corrected_flux_groups = (
            self.hadron_systematics.ppfx_corrected_flux.groupby(level=three_levels)
        )

        flux_prediction_groups = self.flux_prediction.groupby(level=levels)

        product_dict |= convert_groups_to_dict(
            df_groups=stat_uncrt_abs_groups,
            bins=self.bin_edges,
            hist_name_builder=lambda horn, nu: f"hstat_{horn}_{nu}_abs",
            hist_title=";E_{#nu} [GeV];#sigma^{stat} [m^{-2} POT^{-1}]",
            directory_builder=lambda _: "statistical_uncertainties",
        )

        product_dict |= convert_groups_to_dict(
            df_groups=stat_uncrt_frac_groups,
            bins=self.bin_edges,
            hist_name_builder=lambda horn, nu: f"hstat_{horn}_{nu}",
            hist_title=";E_{#nu} [GeV];#sigma^{stat} / #Phi",
            directory_builder=lambda _: "statistical_uncertainties",
        )

        product_dict |= convert_groups_to_dict(
            df_groups=had_uncrt_groups,
            bins=self.bin_edges,
            hist_name_builder=lambda cat, horn, nu: f"hfrac_hadron_{cat}_{horn}_{nu}",
            hist_title=";E_{#nu} [GeV];Fractional Uncertainty",
            directory_builder=lambda cat: f"fractional_uncertainties/hadron/{cat}",
        )

        product_dict |= convert_groups_to_dict(
            df_groups=flux_weights_groups,
            bins=self.bin_edges,
            hist_name_builder=lambda horn, nu: f"hweights_{horn}_{nu}",
            hist_title=";E_{#nu} [GeV]; #Phi_{#nu}^{PPFX} / #Phi_{#nu}^{nom}",
            directory_builder=lambda _: "ppfx_flux_weights",
        )

        product_dict |= convert_groups_to_dict(
            df_groups=ppfx_corrected_flux_groups,
            bins=self.bin_edges,
            hist_name_builder=lambda cat, horn, nu: f"h{cat}_{horn}_{nu}",
            hist_title=";E_{#nu} [GeV]; #phi_{#nu} [m^{-2} POT^{-1}]",
            directory_builder=lambda cat: f"ppfx_corrected_flux/{cat}",
            has_uncerts=True,
        )

        product_dict |= convert_groups_to_dict(
            df_groups=flux_prediction_groups,
            bins=self.bin_edges,
            hist_name_builder=lambda horn, nu: f"hflux_{horn}_{nu}",
            hist_title=";E_{#nu} [GeV]; #phi_{#nu} [m^{-2} POT^{-1}]",
            directory_builder=lambda _: "flux_prediction",
            has_uncerts=True,
        )

        matrix_objects = [
            (
                self.hadron_systematics.covariance_matrices.loc["fractional"],
                lambda x: f"covariance_matrices/hadron/{x}/hcov_{x}",
            ),
            (
                self.hadron_systematics.covariance_matrices.loc["absolute"],
                lambda x: f"covariance_matrices/hadron/{x}/hcov_{x}_abs",
            ),
            (
                self.hadron_systematics.correlation_matrices,
                lambda x: f"covariance_matrices/hadron/{x}/hcorr_{x}",
            ),
            (self.total_covariance_matrix, "hcov_total_abs"),
            (self.total_correlation_matrix, "hcorr_total"),
        ]

        if self.beam_systematics_is_initialized:
            beam_uncrt_groups = self.beam_systematics.fractional_uncertainties.groupby(
                level=three_levels
            )

            beam_shift_groups = self.beam_systematics.beam_systematic_shifts.loc[
                "fractional"
            ].groupby(level=three_levels)

            product_dict |= convert_groups_to_dict(
                df_groups=beam_uncrt_groups,
                bins=self.bin_edges,
                hist_name_builder=lambda cat, horn, nu: f"hfrac_beam_{cat}_{horn}_{nu}",
                hist_title=";E_{#nu} [GeV];Fractional Uncertainty",
                directory_builder=lambda cat: f"fractional_uncertainties/beam/{cat}",
            )

            product_dict |= convert_groups_to_dict(
                df_groups=beam_shift_groups,
                bins=self.bin_edges,
                hist_name_builder=lambda cat, horn, nu: f"hsyst_beam_{cat}_{horn}_{nu}",
                hist_title=";E_{#nu} [GeV]; #phi_{x} - #phi_{nom} / #phi_{nom}",
                directory_builder=lambda _: "beam_systematic_shifts",
            )

            matrix_objects += [
                (
                    self.beam_systematics.covariance_matrices.loc["fractional"],
                    lambda x: f"covariance_matrices/beam/run_{x}/hcov_{x}",
                ),
                (
                    self.beam_systematics.covariance_matrices.loc["absolute"],
                    lambda x: f"covariance_matrices/beam/run_{x}/hcov_{x}_abs",
                ),
                (
                    self.beam_systematics.correlation_matrices,
                    lambda x: f"covariance_matrices/beam/run_{x}/hcorr_{x}",
                ),
                (
                    self.beam_systematics.total_covariance_matrix.loc["fractional"],
                    "covariance_matrices/beam/hcov_total",
                ),
                (
                    self.beam_systematics.total_covariance_matrix.loc["absolute"],
                    "covariance_matrices/beam/hcov_total_abs",
                ),
                (
                    self.beam_systematics.total_correlation_matrix,
                    "covariance_matrices/beam/hcorr_total",
                ),
            ]

        for mat, title_gen in matrix_objects:
            product_dict |= self._export_matrices(mat, title_gen)

        product_dict["pca/hcov_pca"] = convert_pandas_to_th2(
            self.pca_covariance_matrix,
            hist_name="hcov_pca",
        )
        product_dict["pca/cov_pca"] = convert_symmetric_ndarray_to_tmatrix(
            self.pca_covariance_matrix.values,
        )

        eigenvals = convert_pandas_to_th1(
            series=self.pca_eigen_values["eigenvalue"],
            bin_edges=np.arange(len(self.pca_eigen_values["eigenvalue"]), dtype=float),
            hist_name="heigenvals",
        )

        eigenvals.SetTitle(";Principal Component;#lambda_{n}")

        product_dict["pca/heigenvals"] = eigenvals

        frac_eigenvals = convert_pandas_to_th1(
            series=self.pca_eigen_values["fractional_eigenvalue"],
            bin_edges=np.arange(
                len(self.pca_eigen_values["fractional_eigenvalue"]), dtype=float
            ),
            hist_name="heigenvals_frac",
        )

        frac_eigenvals.SetTitle(
            ";Principal Component;#lambda_{n} / #sum_{k} #lambda_{k}"
        )

        product_dict["pca/heigenvals_frac"] = frac_eigenvals

        cumulative_sum = convert_pandas_to_th1(
            series=self.pca_eigen_values["cumulative_sum"],
            bin_edges=np.arange(
                len(self.pca_eigen_values["cumulative_sum"]), dtype=float
            ),
            hist_name="heigenvals_cumulative_sum",
        )

        cumulative_sum.SetTitle(
            ";Principal Component; CUSUM #left(#lambda_{n} / #sum_{k} #lambda_{k}#right)"
        )

        product_dict["pca/heigenvals_cumulative_sum"] = cumulative_sum

        pca_components = self.pca_components.groupby(
            level=("scale", "horn_polarity", "neutrino_mode")
        )

        for (scale, horn, nu), pcs in pca_components:  # type: ignore
            for npc, pc in pcs.items():
                if scale == "evec":
                    hist_title = f"hevec_{npc}_{horn}_{nu}"
                    axis_titles = ";E_{#nu} [GeV]; #hat{e}_{n}"
                    subdir = "eigenvectors"
                else:
                    hist_title = f"hpc_{npc}_{horn}_{nu}"
                    axis_titles = ";E_{#nu} [GeV]; #sqrt{#lambda_{n}} #hat{e}_{n}"
                    subdir = "principal_components"

                th1 = convert_pandas_to_th1(
                    series=pc,
                    bin_edges=self.bin_edges[nu],
                    hist_name=hist_title,
                )

                th1.SetTitle(axis_titles)
                product_dict[f"pca/{subdir}/{hist_title}"] = th1

        return product_dict



import logging
from itertools import product
from pathlib import Path
from typing import Generator, Optional

from hist.hist import Hist
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy import log10
from numpy.typing import NDArray
from ROOT import TH1D  # type: ignore

from flux_tool.vis_scripts.style import ylabel_flux


def get_all_neutrinos() -> Generator[tuple[str, str], None, None]:
    yield from product(["fhc", "rhc"], ["nue", "nuebar", "numu", "numubar"])


def save_figure(
    fig: Figure, fig_name: str, output_dir: Path, tex_caption: str, tex_label: str
) -> None:
    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    for ext in ["png", "pdf"]:
        file_name = f"{output_dir}/{fig_name}.{ext}"
        logging.debug(f"Saving image {file_name}...")
        fig.savefig(file_name, bbox_inches="tight")

    tex_figure = build_latex_figure(f"{fig_name}.pdf", tex_caption, tex_label)

    tex_filename = f"{output_dir}/{fig_name}.tex"

    logging.debug(f"Writing figure to {tex_filename}...")

    with open(tex_filename, "w") as texfile:
        texfile.write(tex_figure)


def build_latex_figure(image_path: str, caption: str, label: str) -> str:
    lines: list[str] = [
        r"\begin{figure}",
        r"    \centering",
        f"    \\includegraphics[width=\\textwidth]{{{image_path}}}",
        f"    \\caption{{{caption}}}",
        f"    \\label{{fig:{label}}}",
        r"\end{figure}",
    ]
    fig: str = "\n".join(lines)

    return fig


def get_hist_scale(h: NDArray | Hist) -> int:
    if isinstance(h, Hist):
        hmax = h.counts().max()
    else:
        hmax = h.max()
    power = log10(hmax)
    return int(round(power))


def create_ylabel_with_scale(scale_factor: int) -> str:
    prefix, units = ylabel_flux.split(" ", 1)
    return prefix + f" $\\mathrm{{\\times 10^{{{scale_factor}}}}}$ " + units


def scale_hist(histogram: NDArray, errors: Optional[NDArray] = None, scale_factor=None):
    if scale_factor is None:
        scale_factor = -1 * get_hist_scale(histogram)
    scaled_histogram = (10**scale_factor) * histogram
    # scaled_ylabel = insert_yscale(scale_factor)

    if errors is not None:
        scaled_errors = (10**scale_factor) * errors
        return scaled_histogram, scaled_errors
    return scaled_histogram


def make_legend_no_errorbars(ax: Axes, **kwargs) -> None:
    handles, labels = ax.get_legend_handles_labels()
    handles = [h[0] for h in handles]  # type: ignore
    ax.legend(handles, labels, **kwargs)


def absolute_uncertainty(total_flux: TH1D, fractional_uncertainty: TH1D) -> TH1D:
    return total_flux * fractional_uncertainty



from functools import partial
from itertools import product
from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np

from flux_tool.vis_scripts.helper import create_ylabel_with_scale, save_figure
from flux_tool.vis_scripts.spectra_reader import SpectraReader
from flux_tool.vis_scripts.style import (
    colorscheme,
    neutrino_labels,
    xlabel_enu,
    ylabel_flux,
)


def plot_flux_prediction(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    xlim: tuple[float, float] = (0, 20),
    label_drawer: Optional[partial] = None,
    bullets: bool = False,
):
    flux_prediction = reader.flux_prediction

    for horn, nu in product(reader.horn_current, ["nue", "numu"]):
        hist_title = f"hflux_{horn}_{nu}"
        nu_flux = flux_prediction[hist_title]
        nubar_flux = flux_prediction[f"{hist_title}bar"]

        flux = [nu_flux.to_pyroot(), nubar_flux.to_pyroot()]

        bins = nu_flux.to_numpy()[1]

        max_flux = flux[0].GetMaximum()
        power = -1 * np.round(np.log10(max_flux))
        scale_factor = 10**power

        nominal = [
            reader[f"beam_samples/run_nominal/hnom_{horn}_{nu}"].to_pyroot(),  # type: ignore
            reader[f"beam_samples/run_nominal/hnom_{horn}_{nu}bar"].to_pyroot(),  # type: ignore
        ]

        for h in flux:
            h.Scale(scale_factor)

        for h in nominal:
            h.Scale(scale_factor)

        ylabel = create_ylabel_with_scale(int(power))

        prediction_label = (
            # "PPFX Mean "
            r"$\left(\pm"
            r"\mathrm{\sigma}_\mathsf{stat}"
            r"\oplus"
            r"\mathrm{\sigma}_\mathsf{syst}"
            r"\right)$"
        )
        prediction_labels = [
            f"Corrected {neutrino_labels[nu]} Flux {prediction_label}",
            f"Corrected {neutrino_labels[f'{nu}bar']} Flux {prediction_label}",
        ]

        nominal_labels = [
            f"Uncorrected {neutrino_labels[nu]} Flux",
            f"Uncorrected {neutrino_labels[f'{nu}bar']} Flux",
        ]

        fig, ax = plt.subplots(figsize=(12,12))

        ax.set_box_aspect(1)

        marker = "o" if nu == "numu" else "s"

        color = [colorscheme["blue"], colorscheme["vermillion"]]
        if nu == "nue":
            color = [colorscheme["bluishgreen"], colorscheme["reddishpurple"]]

        if bullets:
            hep.histplot(
                H=flux,
                label=prediction_labels,
                ax=ax,
                histtype="errorbar",
                binwnorm=True,
                elinewidth=3,
                capsize=4,
                markersize=14,
                markerfacecolor=[None, "none"],
                color=color,
                marker=marker,
            )
        else:
            hep.histplot(
                H=flux,
                label=prediction_labels,
                ax=ax,
                histtype="step",
                binwnorm=True,
                lw=2,
                color=color,
                yerr=False,
                edges=False,
            )
            fill = ["C0", "C1"] if nu == "numu" else ["C2", "C3"]
            for j, f in enumerate(flux):
                err_up = [0.0] + [
                    (f.GetBinContent(i) + f.GetBinError(i)) / f.GetBinWidth(i)
                    for i in range(1, f.GetNbinsX() + 1)
                ]
                err_low = [0.0] + [
                    (f.GetBinContent(i) - f.GetBinError(i)) / f.GetBinWidth(i)
                    for i in range(1, f.GetNbinsX() + 1)
                ]
                ax.fill_between(
                    bins, err_low, err_up, step="pre", color=fill[j], alpha=0.45
                )

        hep.histplot(
            H=nominal,
            label=nominal_labels,
            binwnorm=True,
            ax=ax,
            yerr=False,
            histtype="step",
            color=["k", "gray"],
            ls=["-", "--"],
            lw=2,
            edges=False,
        )

        if label_drawer is not None:
            label_drawer(ax=ax)

        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel_enu)
        ax.legend(loc="best", fontsize=23)

        ax.set_xlim(*xlim)

        if output_dir is not None:
            prefix = f"{horn}_{nu}"
            file_stem = f"{prefix}_flux_prediction"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)

        plt.close(fig)


def plot_flux_uncorrected_logarithmic(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    xlim: tuple[float, float] = (0, 20),
):
    flux_prediction = reader.nominal_spectra
    pot = reader.pot

    for horn in reader.horn_current:
        flux = [
            flux_prediction[f"{horn}/nom/hnom_numu"].to_pyroot(),
            flux_prediction[f"{horn}/nom/hnom_numubar"].to_pyroot(),
            flux_prediction[f"{horn}/nom/hnom_nue"].to_pyroot(),
            flux_prediction[f"{horn}/nom/hnom_nuebar"].to_pyroot(),
        ]

        for f in flux:
            f.Scale(1 / pot[horn])

        ylabel = [r"$\mathrm{\phi_\nu}$", r"$\mathrm{\phi_{\bar{\nu}}}$"]

        ratio_nus = [
            [neutrino_labels["numu"], neutrino_labels["numubar"]],
            [neutrino_labels["nue"], neutrino_labels["nuebar"]],
        ]
        right_sign_numu, wrong_sign_numu = 0, 1
        right_sign_nue, wrong_sign_nue = 2, 3

        if horn == "rhc":
            right_sign_numu, wrong_sign_numu = wrong_sign_numu, right_sign_numu
            right_sign_nue, wrong_sign_nue = wrong_sign_nue, right_sign_nue
            ylabel.reverse()
            ratio_nus = [
                [neutrino_labels["numubar"], neutrino_labels["numu"]],
                [neutrino_labels["nuebar"], neutrino_labels["nue"]],
            ]

        ratio_labels = map("/".join, ratio_nus)

        sign_contam = [flux[right_sign_numu].Clone(), flux[right_sign_nue].Clone()]

        sign_contam[0].Divide(flux[wrong_sign_numu])
        sign_contam[1].Divide(flux[wrong_sign_nue])

        prediction_labels = [
            neutrino_labels["nue"],
            neutrino_labels["nuebar"],
            neutrino_labels["numu"],
            neutrino_labels["numubar"],
        ]

        facecolor = [None, "none", None, "none"]
        markersize = 18
        markeredgewidth = [None, 2, None, 2]

        fig, axs = plt.subplots(
            2,
            1,
            layout="constrained",
            figsize=(12, 12),
            height_ratios=(3, 1),
            sharex=True,
            gridspec_kw={"hspace": 0.03},
        )

        ax1, ax2 = axs

        # ax1.set_box_aspect(1)
        # ax2.set_box_aspect(1)

        hep.histplot(
            ax=ax1,
            H=flux,
            label=prediction_labels,
            histtype="errorbar",
            markerfacecolor=facecolor,
            markersize=markersize,
            markeredgewidth=markeredgewidth,
            marker="o",
            binwnorm=True,
            yerr=False,
        )

        hep.histplot(
            ax=ax2,
            H=sign_contam,
            lw=3,
            label=ratio_labels,
            yerr=False,
            edges=False,
        )

        ax1.legend(loc="upper right", ncol=2)
        ax1.set_yscale("log")
        ax1.set_ylabel(ylabel_flux)

        ax1.set_xlim(xlim)
        ax2.set_xlim(xlim)

        ax2.set_xlabel(xlabel_enu)
        ax2.legend(loc="upper right")
        ax2.axhline(1, ls="--", lw=2, color="k")
        ax2.set_ylabel(" / ".join(ylabel))

        hep.label.exp_label(
            exp="NuMI",
            llabel=f"Simulation ({horn.upper()}, Uncorrected)",
            rlabel="",
            ax=ax1,
        )

        if output_dir is not None:
            file_stem = f"{horn}_uncorrected_flux_log"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)

        plt.close(fig)



import logging
import signal
import sys
from argparse import ArgumentParser
from importlib.util import find_spec
from time import time

from flux_tool.config import AnalysisConfig
from flux_tool.exporter import Exporter
from flux_tool.flux_systematics_analysis import FluxSystematicsAnalysis
from flux_tool.preprocessor import Preprocessor
from flux_tool.vis_scripts.plot_all import compress_directory, plot_all


def timer(fn):
    def wrap():
        start = time()
        fn()
        end = time()
        logging.info(f"Finished in {end-start:0.2f} s")

    return wrap


def signal_handler(sig, frame):
    print("Interrupted. Exiting...")
    exit(0)


signal.signal(signal.SIGINT, signal_handler)


def check_for_ROOT() -> None:
    try:
        spec = find_spec("ROOT")
    except ValueError:
        """For some reason ROOT doesn't have a defined __spec__,
        and a ValueError exception is thrown.
        But it's still importable so we can catch the exception here.
        """
        pass
    else:
        if not spec:
            msg = (
                "PyROOT is not installed."
                "Please install ROOT before using this package."
            )
            raise ImportError(msg)


def load_config(cfg_path: str) -> AnalysisConfig:
    try:
        cfg = AnalysisConfig.from_file(cfg_path)
    except FileNotFoundError:
        print(f'The configuration file "{cfg_path}" was not found. Exiting...')
        sys.exit(1)
    return cfg


def run_analysis(cfg: AnalysisConfig):
    preprocessor = Preprocessor(cfg=cfg)

    analysis = FluxSystematicsAnalysis(
        nominal_flux_df=preprocessor.nominal_flux_df,
        ppfx_correction_df=preprocessor.ppfx_correction_df,
        bin_edges=cfg.bin_edges,
        cfg=cfg,
    )

    analysis.run(pca_threshold=1)

    exporter = Exporter(cfg, analysis)

    exporter.export_products()

    exporter.export_ppfx_output()

    with open(
        exporter.products_file.parent / "flux_covariance_binning_NuMI_GeV.txt", "w"
    ) as f:
        f.write(analysis.matrix_binning_str)

    return exporter.products_file


@timer
def main():
    check_for_ROOT()

    parser = ArgumentParser(
        prog="flux_uncertainties",
        description="This package coerces PPFX output into a neutrino flux prediction with uncertainties, and stores various spectra related to the flux, e.g., fractional uncertainties, covariance matrices, etc.",
    )
    parser.add_argument(
        "-c", "--config", help="specify the path to a toml configuration file"
    )

    parser.add_argument(
        "-p",
        "--plots-only",
        dest="plot",
        metavar="PRODUCTS_FILE",
        help="Specify path to an existing ROOT file for which to produce plots",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        help="",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        default=logging.INFO,
    )

    parser.add_argument(
        "-z",
        "--enable-compression",
        action="store_true",
        dest="compression",
        help="Enable compression of the output plots directory",
    )

    parser.add_argument("--example-config", action="store_true", help="Print an example configuration file")

    args = parser.parse_args()

    logging.basicConfig(format="%(message)s", level=args.loglevel)

    if args.example_config:
        with open("./config.toml", "w", encoding="utf-8") as f:
            f.write(AnalysisConfig.dump_default_config())
        logging.info("Example configuration file written to config.toml")
        sys.exit(0)

    cfg_str = args.config

    if cfg_str is None:
        parser.print_help()
        sys.exit(1)

    cfg = load_config(cfg_str)

    plot = args.plot

    products_file = plot if plot is not None else run_analysis(cfg)

    logging.info("\n=============== MAKING PLOTS ===============")

    plot_all(products_file, cfg.plots_path, cfg.plot_opts, cfg.bin_edges)

    if args.compression:
        compress_directory(cfg.plots_path)

    logging.info("Done.")


if __name__ == "__main__":
    main()



import logging
import tarfile
from functools import partial
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import mplhep as hep
from numpy.typing import NDArray

import flux_tool.vis_scripts as vis
from flux_tool.vis_scripts.spectra_reader import SpectraReader
from flux_tool.vis_scripts.style import style


def plot_all(
    products_file: Path | str,
    output_dir: Path,
    plot_opts: dict[str, Any],
    binning: dict[str, NDArray],
):
    plt.style.use(style)

    reader = SpectraReader(products_file, binning)

    if plot_opts["draw_label"]:
        label_drawer = partial(
            hep.label.exp_label,
            exp=plot_opts["experiment"],
            llabel=plot_opts["stage"],
            rlabel="",
        )
    else:
        label_drawer = None

    xlim: tuple[float, float] = plot_opts["xlim"]

    enabled_plots = plot_opts["enabled"]

    if all(enabled_plots.values()):
        logging.info(f"Pre-caching ROOT objects from {products_file}...")
        reader.load_cache()
        logging.info("Done.")

    if enabled_plots["uncorrected_flux"]:
        logging.info("Plotting uncorrected flux...")
        vis.plot_flux_uncorrected_logarithmic(
            reader, output_dir / "flux_spectra/uncorrected_flux", xlim
        )

    if enabled_plots["flux_prediction"]:
        logging.info("Plotting flux prediction...")
        vis.plot_flux_prediction(
            reader,
            output_dir / "flux_spectra/flux_prediction",
            xlim,
            label_drawer,
            bullets=plot_opts["flux_prediction_bullets"],
        )

    if enabled_plots["flux_prediction_parent_spectra"]:
        logging.info("Plotting parent spectra...")
        vis.plot_parents(reader, output_dir / "flux_spectra/parents", xlim)

    if enabled_plots["flux_prediction_parent_spectra_stacked"]:
        logging.info("Plotting parent spectra (stacked)...")
        vis.plot_parents(
            reader,
            output_dir / "flux_spectra/parents",
            xlim,
            stacked=True,
        )

    if enabled_plots["ppfx_universes"]:
        logging.info("Plotting PPFX universes...")
        vis.plot_ppfx_universes(reader, output_dir / "flux_spectra/universes", xlim)

    if enabled_plots["hadron_uncertainties"]:
        logging.info("Plotting hadron uncertainties...")
        vis.plot_uncertainties(
            reader,
            vis.plot_hadron_fractional_uncertainties,
            output_dir / "hadron_uncertainties",
            xlim,
            label_drawer=label_drawer,
        )

    if enabled_plots["hadron_uncertainties_meson"]:
        logging.info("Plotting hadron uncertainties (mesinc)...")
        vis.plot_uncertainties(
            reader,
            vis.plot_hadron_fractional_uncertainties_mesinc_breakout,
            output_dir / "hadron_uncertainties/meson_breakout",
            xlim,
        )
    if enabled_plots["hadron_uncertainties_meson_only"]:
        logging.info("Plotting hadron uncertainties (mesinc only)...")
        logging.warning("Deprecated: Use hadron_uncertainties_meson instead.")
        # plot_hadron_fractional_uncertainties_mesinc_only(
        #     reader, output_dir / "hadron_uncertainties/meson_only", xlim
        # )

    if enabled_plots["hadron_uncertainties_nua"]:
        logging.info("Plotting hadron uncertainties (nua bands only)...")
        vis.plot_uncertainties(
            reader,
            vis.plot_hadron_fractional_uncertainties_nua_breakout,
            output_dir / "hadron_uncertainties",
            xlim,
        )

    if enabled_plots["pca_scree_plot"]:
        logging.info("Plotting Eigenvalues Scree Plot")
        vis.scree_plot(reader, output_dir / "pca")

    if enabled_plots["pca_mesinc_overlay"]:
        logging.info("Plotting PCA Uncertainty Comparison")
        vis.pca_mesinc_overlay(reader, output_dir / "pca")

    if enabled_plots["pca_top_components"]:
        logging.info("Plotting top principal components")
        vis.plot_top_principal_components(reader, output_dir / "pca")

    if enabled_plots["pca_variances"]:
        logging.info("Plotting PCA variances...")
        vis.plot_hadron_systs_and_pca_variances(
            reader, output_dir / "pca/variances", xlim
        )

    if enabled_plots["pca_components"]:
        logging.info("Plotting princpal components...")
        vis.plot_pca_systematic_shifts(
            reader, output_dir / "pca/components", xlim, (-0.12, 0.12)
        )

    if enabled_plots["beam_uncertainties"]:
        logging.info("Plotting beam uncertainties...")
        vis.plot_uncertainties(
            reader,
            vis.plot_beam_fractional_uncertainties,
            output_dir / "beam_uncertainties",
            xlim,
            (0, 0.12),
        )

    if status := enabled_plots["hadron_covariance_matrices"]:
        logging.info("Plotting hadron covariance matrices...")
        vis.plot_hadron_covariance_matrices(
            reader,
            output_dir / "covariance_matrices/hadron",
            which=status if isinstance(status, str) else None,
        )

    if status := enabled_plots["beam_covariance_matrices"]:
        logging.info("Plotting beamline focusing covariance matrices...")
        vis.plot_beam_covariance_matrices(
            reader,
            output_dir / "covariance_matrices/beam",
            which=status if isinstance(status, str) else None,
        )

    if status := enabled_plots["hadron_correlation_matrices"]:
        logging.info("Plotting hadron correlation matrices...")
        vis.plot_hadron_correlation_matrices(
            reader,
            output_dir / "covariance_matrices/hadron",
            which=status if isinstance(status, str) else None,
        )

    if status := enabled_plots["beam_correlation_matrices"]:
        logging.info("Plotting beamline focusing correlation matrices...")
        vis.plot_beam_correlation_matrices(
            reader,
            output_dir / "covariance_matrices/beam",
            which=status if isinstance(status, str) else None,
        )

    if enabled_plots["beam_systematic_shifts"]:
        logging.info("Plotting beamline systematic shifts...")
        vis.plot_beam_systematic_shifts(
            reader, output_dir / "beam_systematic_shifts", xlim
        )


def compress_directory(directory: Path):
    logging.info(f"Compressing {directory}...")
    with tarfile.open("plots.tar.xz", "w:xz") as tar:
        for d in directory.iterdir():
            tar.add(d, arcname=d.stem)



from functools import cached_property
from itertools import product
from pathlib import Path

from numpy.typing import NDArray
import uproot


class SpectraReader:
    def __init__(self, products_file: Path | str, binning: dict[str, NDArray]):
        self._f = uproot.open(products_file)

        self.binning = binning

        horns_nus = [
            key.split("_")[0].split("/")
            for key in self._f["ppfx_output"].keys(filter_name="*/nu*", cycle=False)  # type: ignore
        ]


        self.horn_current = list({item[0] for item in horns_nus})
        self.neutrinos = list({item[1] for item in horns_nus})

        self.horns_and_nus = list(product(self.horn_current, self.neutrinos))

    def __del__(self):
        self._f.close()  # type: ignore

    def __getitem__(self, key):
        return self._f[key]

    def load_cache(self) -> None:
        self.beam_covariance_matrices
        self.beam_correlation_matrices
        self.beam_systematic_shifts
        self.beam_uncertainties
        self.flux_prediction
        self.flux_weights
        self.hadron_covariance_matrices
        self.hadron_correlation_matrices
        self.hadron_uncertainties
        self.nominal_spectra
        self.parent_spectra
        self.pot
        self.ppfx_correction
        self.pca_eigenvalues
        self.principal_components
        self.universes

    @cached_property
    def beam_uncertainties(self):
        return {
            key: h
            for key, h in self._f["fractional_uncertainties/beam/"].items(
                cycle=False, filter_classname="TH1D"  # type: ignore
            )
        }

    @cached_property
    def beam_systematic_shifts(self):
        return {key: h for key, h in self._f["beam_systematic_shifts"].items(cycle=False)}  # type: ignore

    @cached_property
    def flux_prediction(self):
        return {key: h for key, h in self._f["flux_prediction"].items(cycle=False)}  # type: ignore

    @cached_property
    def flux_weights(self):
        return {
            key: h for key, h in self._f["ppfx_flux_weights"].items(cycle=False)  # type: ignore
        }

    @cached_property
    def hadron_covariance_matrices(self):
        return {
            key: h
            for key, h in self._f["covariance_matrices/hadron"].items(
                filter_name="*/hcov*", filter_classname="TH2D", cycle=False  # type: ignore
            )
        }

    @cached_property
    def hadron_correlation_matrices(self):
        return {
            key: h
            for key, h in self._f["covariance_matrices/hadron"].items(
                filter_name="*/hcor*", filter_classname="TH2D", cycle=False  # type: ignore
            )
        }

    @cached_property
    def beam_covariance_matrices(self):
        return {
            key: h
            for key, h in self._f["covariance_matrices/beam"].items(
                filter_name="*hcov*", filter_classname="TH2D", cycle=False  # type: ignore
            )
        }

    @cached_property
    def beam_correlation_matrices(self):
        return {
            key: h
            for key, h in self._f["covariance_matrices/beam"].items(
                filter_name="*hcor*", filter_classname="TH2D", cycle=False  # type: ignore
            )
        }

    @cached_property
    def hadron_uncertainties(self):
        return {
            key: h
            for key, h in self._f["fractional_uncertainties/hadron/"].items(
                cycle=False, filter_classname="TH1D"  # type: ignore
            )
        }

    @cached_property
    def nominal_spectra(self):
        return {
            key: h
            for key, h in self._f["ppfx_output"].items(
                cycle=False, filter_name="*/nom/hnom*"  # type: ignore
            )
        }

    @cached_property
    def parent_spectra(self):
        return {
            key: h
            for key, h in self._f["ppfx_output"].items(
                cycle=False, filter_name="*/nom/parent/*"  # type: ignore
            )
        }

    @cached_property
    def pot(self):
        pot = {
            horn: self._f[f"ppfx_output/{horn}/hpot"].values().max()  # type: ignore
            for horn in self.horn_current
        }
        return pot

    @cached_property
    def ppfx_correction(self):
        return {
            key: h for key, h in self._f["ppfx_corrected_flux/total"].items(cycle=False)  # type: ignore
        }

    @cached_property
    def principal_components(self):
        return {
            key: h
            for key, h in self._f["pca/principal_components"].items(
                cycle=False, filter_name="*hpc_*"  # type: ignore
            )
        }

    @cached_property
    def pca_eigenvalues(self):
        return self._f["pca/heigenvals_frac"]

    @cached_property
    def universes(self):
        return {
            key: h
            for key, h in self._f["ppfx_output"].items(
                filter_name="*/*_total/*", cycle=False  # type: ignore
            )
        }



from itertools import cycle

import matplotlib.pyplot as plt
import mplhep
from cycler import Cycler, cycler
from matplotlib.axes import Axes

# Okabe-Ito Colors
colorscheme = {
    "blue": "#0072B2",
    "vermillion": "#D55E00",
    "bluishgreen": "#009E73",
    "reddishpurple": "#CC79A7",
    "skyblue": "#56B4E9",
    "orange": "#E69F00",
    "yellow": "#F0E442",
}

style = {
    "axes.formatter.limits": [-5, 6],
    "axes.formatter.use_mathtext": True,
    "axes.labelsize": 28,
    "axes.linewidth": 1.0,
    "axes.prop_cycle": cycler(linestyle=["-", "--", ":", "-."])
    * cycler(
        color=[
            "#0072B2",
            "#D55E00",
            "#009E73",
            "#CC79A7",
            "#56B4E9",
            "#E69F00",
            "#F0E442",
            # "#000000",
        ]
    ),
    "axes.unicode_minus": False,
    "axes.titlesize": 32,
    "figure.figsize": [10.0, 10.0],
    # "font.sans-serif": "TeX Gyre Heros",
    "font.family": "sans-serif",
    "font.size": 22,
    "grid.alpha": 0.8,
    "grid.linestyle": "--",
    "grid.linewidth": 0.5,
    "image.cmap": "cividis",
    "lines.linewidth": 2.0,
    "legend.fontsize": 28,
    "legend.handlelength": 1.5,
    "legend.borderpad": 0.5,
    "legend.frameon": False,
    "mathtext.default": "rm",
    "mathtext.fontset": "stix",
    "savefig.transparent": True,
    "xaxis.labellocation": "center",
    "yaxis.labellocation": "center",
    "xtick.labelsize": "small",
    "xtick.direction": "in",
    "xtick.major.size": 12,
    "xtick.minor.size": 6,
    "xtick.major.pad": 10,
    "xtick.top": True,
    "xtick.major.top": True,
    "xtick.major.bottom": True,
    "xtick.minor.top": True,
    "xtick.minor.bottom": True,
    "xtick.minor.visible": True,
    "ytick.labelsize": "small",
    "ytick.direction": "in",
    "ytick.major.size": 12,
    "ytick.minor.size": 6.0,
    "ytick.right": True,
    "ytick.major.left": True,
    "ytick.major.right": True,
    "ytick.minor.left": True,
    "ytick.minor.right": True,
    "ytick.minor.visible": True,
}

okabe_ito = {
    "blue": "#0072B2",
    "vermillion": "#D55E00",
    "bluishgreen": "#009E73",
    "reddishpurple": "#CC79A7",
    "skyblue": "#56B4E9",
    "orange": "#E69F00",
    "black": "#000000",
    "yellow": "#F0E442",
}

neutrino_labels = {
    "nue": r"$\mathrm{\nu_{e}}$",
    "nuebar": r"$\mathrm{\bar{\nu}_{e}}$",
    "numu": r"$\mathrm{\nu_{\mu}}$",
    "numubar": r"$\mathrm{\bar{\nu}_{\mu}}$",
}

neutrino_parent_labels = {
    "kp": r"$\mathrm{K^{+}}$",
    "km": r"$\mathrm{K^{-}}$",
    "kpm": r"$\mathrm{K^{\pm}}$",
    "k0l": r"$\mathrm{K^{0}_{L}}$",
    "mup": r"$\mathrm{\mu^{+}}$",
    "mum": r"$\mathrm{\mu^{-}}$",
    "mu": r"$\mathrm{\mu^{\pm}}$",
    "pipm": r"$\mathrm{\pi^{\pm}}$",
    "pip": r"$\mathrm{\pi^{+}}$",
    "pim": r"$\mathrm{\pi^{-}}$",
}

ppfx_labels = {
    "pCpi": r"$\mathrm{p + C \to \pi^\pm + X}$",
    "pCnu": r"$\mathrm{p + C \to N + X}$",
    "pCk": r"$\mathrm{p + C \to K + X}$",
    "pCQEL": r"$\mathrm{p + C}$ (QEL)",
    "nCpi": r"$\mathrm{n + C \to \pi^\pm + X}$",
    "mesinc_projectile_K0": r"$\mathrm{K^0 + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_Km": r"$\mathrm{K^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_Kp": r"$\mathrm{K^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_pip": r"$\mathrm{\pi^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_pim": r"$\mathrm{\pi^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_K0": r"$\mathrm{K^0 + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_Km": r"$\mathrm{K^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_Kp": r"$\mathrm{K^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_pip": r"$\mathrm{\pi^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_pim": r"$\mathrm{\pi^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_daughter_K0": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^0 + X}$",
    "mesinc_daughter_Km": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{-} + X}$",
    "mesinc_daughter_Kp": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{+} + X}$",
    "mesinc_daughter_pip": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{+} + X}$",
    "mesinc_daughter_pim": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{-} + X}$",
    "mesinc_outgoing_K0": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^0 + X}$",
    "mesinc_outgoing_Km": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{-} + X}$",
    "mesinc_outgoing_Kp": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{+} + X}$",
    "mesinc_outgoing_pip": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{+} + X}$",
    "mesinc_outgoing_pim": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{-} + X}$",
    "mesinc": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \left( \pi^\pm, K, N \right) + X}$",
    "nua": r"$\mathrm{N + A \to X}$",
    "nua_datavol": r"$\mathrm{N + C}$ OOPS ($\mathrm{-0.25 \leq x_F < 0}$)",
    "nua_inC_inPS": r"$\mathrm{N + C}$ OOPS ($\mathrm{-0.25 \leq x_F < 0}$)",
    "nua_datavol_negxF": r"$\mathrm{N + C}$ OOPS ($\mathrm{x_F < -0.25}$)",
    "nua_inC_OOPS": r"$\mathrm{N + C}$ OOPS ($\mathrm{x_F < -0.25}$)",
    "nua_othervol": r"$\mathrm{N + A}$ in PS",
    "nua_outC_Ascale": r"$\mathrm{N + A}$ in PS",
    "nua_outC_OOPS": r"$\mathrm{N + A}$ OOPS",
    # "nua_other": r"$\mathrm{N + A}$ Other",
    "nua_other": r"$\mathrm{N + C}$ Other",
    "nuAlFe": r"$\mathrm{N + (Al, Fe)\to X}$",
    "att": "Attenuation",
    "others": "Others",
    "total": "Total",
}

ppfx_mesinc_labels = {
    "mesinc_projectile_K0": r"$\mathrm{K^0 + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_Km": r"$\mathrm{K^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_Kp": r"$\mathrm{K^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_pip": r"$\mathrm{\pi^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_projectile_pim": r"$\mathrm{\pi^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_K0": r"$\mathrm{K^0 + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_Km": r"$\mathrm{K^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_Kp": r"$\mathrm{K^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_pip": r"$\mathrm{\pi^{+} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_incoming_pim": r"$\mathrm{\pi^{-} + A \to \left( \pi^\pm, K, N \right) + X}$",
    "mesinc_daughter_K0": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^0 + X}$",
    "mesinc_daughter_Km": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{-} + X}$",
    "mesinc_daughter_Kp": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{+} + X}$",
    "mesinc_daughter_pip": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{+} + X}$",
    "mesinc_daughter_pim": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{-} + X}$",
    "mesinc_outgoing_K0": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^0 + X}$",
    "mesinc_outgoing_Km": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{-} + X}$",
    "mesinc_outgoing_Kp": r"$\mathrm{\left( \pi^\pm, K \right) + A \to K^{+} + X}$",
    "mesinc_outgoing_pip": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{+} + X}$",
    "mesinc_outgoing_pim": r"$\mathrm{\left( \pi^\pm, K \right) + A \to \pi^{-} + X}$",
}

ppfx_colors = {
    k: c
    for k, c in zip(ppfx_labels, cycle(colorscheme.values()))
    if "projectile" not in k and "daughter" not in k and "incoming" not in k and "outgoing" not in k
}
ppfx_mesinc_colors = {
    k: c
    for k, c in zip(ppfx_mesinc_labels, cycle(colorscheme.values()))
    if "mesinc" in k
}

ppfx_lines = {
    k: ls
    for k, ls in zip(ppfx_labels, cycle(["-", "--", ":", "-."]))
    if "projectile" not in k and "daughter" not in k and "incoming" not in k and "outgoing" not in k
}

beam_syst_labels = {
    "beam_shift_y_plus": r"Beam Shift $y+1$ mm",
    "beam_shift_y_minus": r"Beam Shift $y-1$ mm",
    "beam_shift_y": r"Beam Shift $y \pm 1$ mm",
    "beam_shift_x_plus": r"Beam Shift $x+1$ mm",
    "beam_shift_x_minus": r"Beam Shift $x-1$ mm",
    "beam_shift_x": r"Beam Shift $x \pm 1$ mm",
    "beam_spot": r"Beam Spot Size $\pm 0.2$ cm",
    "horn_current_plus": "Horn Current $+ 2$ kA",
    "horn_current_minus": "Horn Current $- 2$ kA",
    "horn_current": r"Horn Current $\pm 2$ kA",
    "horn1_y": r"Horn 1 y-Position $\pm 0.3$ cm",
    "horn1_x": r"Horn 1 x-Position $\pm 0.3$ cm",
    "water_layer": r"Horn Water Layer $\pm 1$ mm",
    "beam_div": r"54 $\mathrm{\mu}$rad Beam Divergence",
    "total": "Total",
}

beam_syst_colors = {k: c for k, c in zip(beam_syst_labels, cycle(colorscheme.values()))}

beam_syst_lines = {
    k: ls
    for k, ls in zip(beam_syst_labels, cycle(["-", "--", ":"]))
    if "total" not in k
}


xlabel_enu = r"E$_\mathrm{\nu}$ [GeV]"
ylabel_flux = r"$\mathrm{\phi_\nu}$  / GeV / POT"


def get_prop_cycler() -> Cycler:
    """Returns an axis prop cycler, which cycles first through solid lines with the Okabe Ito colorscheme, then dashed lines.
    Excludes the black color.
    """
    prop_cycle = cycler(linestyle=["-", "--", "-."]) * cycler(
        color=[v for k, v in okabe_ito.items() if k != "black"]
    )

    return prop_cycle


def place_header(
    ax: Axes, header: str, xy: tuple[float, float] = (0.0, 1.0), ha="left", **kwargs
) -> None:
    ax.annotate(
        header,
        xy,
        xytext=(0, 3),
        xycoords="axes fraction",
        textcoords="offset points",
        ha=ha,
        va="bottom",
        fontweight="bold",
        **kwargs,
    )


def icarus_preliminary(
    ax: Axes, xy: tuple[float, float] = (0.0, 1.0), **kwargs
) -> None:
    """
    Annotates a Matplotlib axes object with the label "ICARUS Preliminary".

    This function adds a text annotation to the specified axes at the given
    coordinates (xy). The text "ICARUS Preliminary" is used as the annotation,
    and it is positioned with an offset from the specified coordinates.

    Parameters:
    ax (Axes): The Matplotlib axes object to which the annotation will be added.
    xy (tuple[float, float], optional): The coordinates (x, y) where the annotation
        arrow will point. Default is (0.0, 1.0).
    **kwargs: Additional keyword arguments that are passed to the `annotate` function
        of Matplotlib.

    Returns:
    None

    Example:
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    ax.plot([0, 1], [0, 1])
    icarus_preliminary(ax)  # Annotate with "ICARUS Preliminary" label
    plt.show()
    """
    ax.annotate(
        "ICARUS Preliminary",
        xy,
        xytext=(0, 3),
        xycoords="axes fraction",
        textcoords="offset points",
        ha="left",
        va="bottom",
        fontweight="bold",
        **kwargs,
    )


def apply_style() -> None:
    plt.style.use(mplhep.style.ROOT)

    default_cycler = cycler(linestyle=["-", "--", "-."]) * cycler(
        color=list(okabe_ito.values())
    )

    mpl_opts = {
        "axes.prop_cycle": default_cycler,
        "axes.formatter.use_mathtext": True,
        "axes.formatter.limits": [-3, 3],
        "xaxis.labellocation": "center",
        "yaxis.labellocation": "center",
        "font.family": "TeX Gyre Heros",
        "font.size": 32,
        "mathtext.fontset": "stixsans",
    }

    plt.rcParams.update(mpl_opts)



from typing import Any, Callable, Optional
from uuid import uuid4

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy
from ROOT import TH1D, TH2D, TMatrixDSym  # type: ignore


def get_bin_edges_from_dataframe(df: pd.DataFrame) -> np.ndarray:
    energy_columns = ["E_low", "E_high"]
    if not set(energy_columns).issubset(df.columns):
        raise ValueError(
            "DataFrame does not contain the lower and upper energy bin edges."
        )
    return np.unique(df[energy_columns].to_numpy().flatten())


def calculate_correlation_matrix(
    covariance_matrix: pd.DataFrame | NDArray[Any],
) -> pd.DataFrame | NDArray[Any]:
    """Calculates the correlation matrix for a given covariance matrix."""
    variance = np.sqrt(np.diag(covariance_matrix))
    outer_product = np.outer(variance, variance)
    correlation_matrix = np.divide(
        covariance_matrix,
        outer_product,
        out=np.zeros(covariance_matrix.shape),
        where=outer_product != 0,
    )
    return correlation_matrix

def rebin_within_xlim(hist: TH1D, binning: NDArray, xlim: tuple[float, float]) -> TH1D:
    new_binning = binning[(binning >= xlim[0]) & (binning <= xlim[1])]
    hist = hist.Rebin(len(new_binning) - 1, hist.GetName(), new_binning)
    return hist

def convert_pandas_to_th1(
    series: pd.Series | pd.DataFrame,
    bin_edges: np.ndarray,
    hist_name: Optional[str] = None,
    hist_title: str = "",
    uncerts: Optional[pd.Series] = None,
) -> TH1D:
    if hist_name is None:
        hist_name = str(uuid4())

    th1 = TH1D(hist_name, hist_title, len(bin_edges) - 1, bin_edges)

    if uncerts is None:
        for b, val in enumerate(series.values):
            th1.SetBinContent(b + 1, val)
    else:
        for b, (val, unc) in enumerate(zip(series.values, uncerts.values)):
            th1.SetBinContent(b + 1, val)
            th1.SetBinError(b + 1, unc)

    return th1


def convert_pandas_to_th2(dataframe: pd.DataFrame, hist_name: str) -> TH2D:
    rows = dataframe.index
    columns = dataframe.columns

    xbins = np.arange(0, len(rows) + 1, dtype=float)
    ybins = np.arange(0, len(columns) + 1, dtype=float)
    nbinsx = xbins.shape[0] - 1
    nbinsy = ybins.shape[0] - 1

    th2 = TH2D(hist_name, "", nbinsx, xbins, nbinsy, ybins)

    for ii, (_, row) in enumerate(dataframe.iterrows()):
        for jj, (_, col) in enumerate(row.items()):
            th2.SetBinContent(ii + 1, jj + 1, col)

    it = zip(enumerate(rows), enumerate(columns))

    for (ii, row), (jj, col) in it:
        labelx = f"{row[0]}-{row[1]}-{row[2]}"
        labely = f"{col[0]}-{col[1]}-{col[2]}"
        th2.GetXaxis().SetBinLabel(ii + 1, labelx)
        th2.GetYaxis().SetBinLabel(jj + 1, labely)

    return th2


def convert_symmetric_ndarray_to_tmatrix(matrix: NDArray) -> TMatrixDSym:
    m = TMatrixDSym(matrix.shape[0])
    for i, x in enumerate(matrix):
        for j, y in enumerate(x):
            if i == j:
                m[i, i] = y
            elif i < j:
                m[i, j] = y
                m[j, i] = y
    return m


def convert_groups_to_dict(
    df_groups: SeriesGroupBy | DataFrameGroupBy,
    bins: dict[str, NDArray],
    hist_name_builder: Callable[..., str],
    hist_title: str,
    directory_builder: Callable[[str], str] = lambda _: "",
    has_uncerts: bool = False,
) -> dict[str, TH1D]:
    out_dict: dict[str, TH1D] = {}

    for idx, hist in df_groups:
        nu = idx[-1]  # type: ignore
        hist_name = hist_name_builder(*idx)  # type: ignore
        if has_uncerts:
            th1 = convert_pandas_to_th1(
                hist.iloc[:, 0], bins[nu], hist_name, hist_title, hist.iloc[:, 1]
            )
        else:
            th1 = convert_pandas_to_th1(hist, bins[nu], hist_name, hist_title)
        out_dict[f"{directory_builder(idx[0])}/{hist_name}"] = th1  # type: ignore

    return out_dict



import numpy as np
import pandas as pd


class PCA:
    """Dataclass to carry out a principal component analysis.

    Attributes:
        threshold : float
            Threshold value for the cumulative sum of eigenvalues, above which components are discarded.
            Defaults to 2.0, which keeps all components.
        eigenvectors : np.ndarray
            2-D array containing all eigenvectors as rows, sorted by eigenvalue.
        eigenvalues : np.ndarray
            Array containing the list of eigenvalues in descending order.
        fractional_eigenvalues : np.ndarray
            Array containing the fraction of the variance of the dataset explained by the eigenvalues.
        cumulative_sum : np.ndarray
            Array containing the cumulative sums of fractional_eigenvalues.
        eigenvalues_df : pd.DataFrame
            Pandas DataFrame representation of eigenvalues.
            Also contains columns for fractional_eigenvalues and cumulative_sum.
        principal_component_df : pd.DataFrame
            Pandas DataFrame representation of eigenvectors weighted by eigenvalue.
        new_covariance_matrix: np.ndarray
            Array representing the new covariance matrix reconstructed from the principal components.
    """

    __slots__ = (
        "covariance_matrix",
        "threshold",
        "eigenvalues",
        "eigenvectors",
        "eigenvalues_df",
        "principal_component_df",
    )

    def __init__(self, covariance_matrix: pd.DataFrame, threshold: float = 2) -> None:
        self.covariance_matrix = covariance_matrix
        self.threshold = threshold

    def fit(self) -> None:
        """Perform eigendecomposition of covariance_matrix and sort resulting eigenvectors by largest eigenvalues.

        Parameters:
                covariance_matrix: np.ndarray | pd.DataFrame
        """
        eigenvalues, eigenvectors = np.linalg.eigh(self.covariance_matrix)

        # reversing the order so that largest eigenvalue comes first
        eigenvalues, eigenvectors = eigenvalues[::-1], eigenvectors[:, ::-1]

        # DEBUG
        # quantify magnitude of negative eigenvalues
        # max_eval = max(eigenvalues)
        # min_eval = min(eigenvalues)
        # print(f"|{min_eval}|/{max_eval}={np.abs(min_eval)/max_eval}")

        eigenvectors = eigenvectors[:, eigenvalues > 0]
        eigenvalues = eigenvalues[eigenvalues > 0]

        fractional_eigenvalues = eigenvalues / eigenvalues.sum()

        cumulative_sum = np.cumsum(fractional_eigenvalues)

        selected_components = cumulative_sum <= self.threshold

        self.eigenvectors = eigenvectors[:, selected_components]
        self.eigenvalues = eigenvalues[selected_components]

        index = self.covariance_matrix.index

        self.principal_component_df = pd.concat(
            [
                pd.DataFrame(
                    self.eigenvectors,
                    index=index,
                ),
                pd.DataFrame(
                    np.sqrt(self.eigenvalues) * self.eigenvectors,
                    index=index,
                ),
            ],
            keys=["evec", "evec_scaled"],
            names=["scale"] + index.names,
        )

        self.eigenvalues_df = pd.DataFrame(
            self.eigenvalues,
            columns=["eigenvalue"],
        )

        self.eigenvalues_df.index.name = "principal_component"

        self.eigenvalues_df["fractional_eigenvalue"] = fractional_eigenvalues[
            selected_components
        ]

        self.eigenvalues_df["cumulative_sum"] = cumulative_sum[selected_components]

    @property
    def new_covariance_matrix(self) -> np.ndarray:
        return self.eigenvectors @ np.diag(self.eigenvalues) @ self.eigenvectors.T



from dataclasses import dataclass, field
from functools import cached_property
from typing import Any, Optional

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from flux_tool.helpers import (calculate_correlation_matrix,
                               convert_pandas_to_th1)


def smooth_stat_fluctuations(
    df: pd.DataFrame, bin_edges: dict[str, np.ndarray]
) -> pd.DataFrame:
    groups = df.stack("run_id").groupby(by=["run_id", "horn_polarity", "neutrino_mode"])

    smoothed_flux_list = []

    for (_, _, nu), g in groups:  # type: ignore
        th1 = convert_pandas_to_th1(series=g, bin_edges=bin_edges[nu])  # type: ignore

        th1.Smooth()

        smoothed_fluxes = [th1[i[-2]] for i in g.index]  # type: ignore

        series = pd.Series(smoothed_fluxes, index=g.index)

        smoothed_flux_list.append(series)

    return pd.concat(smoothed_flux_list).unstack("run_id")  # type: ignore


@dataclass(repr=False)
class BeamFocusingSystematics:
    beam_flux_df: pd.DataFrame
    bin_edges: dict[str, np.ndarray]
    smoothing: Optional[bool] = False
    nominal_run: pd.DataFrame = field(init=False)
    _beam_pt: pd.DataFrame = field(init=False)
    _run_id_map: dict[str, int | tuple[int, int]] = field(init=False)

    def __post_init__(self) -> None:
        self._beam_pt = pd.pivot_table(
            data=self.beam_flux_df.query("category == 'nominal'"),
            values=["flux"],
            index=["horn_polarity", "neutrino_mode", "bin"],
            columns=["run_id"],
        )["flux"]

        self._run_id_map = {
            "beam_power": 1,
            "horn_current_plus": 8,
            "horn1_x": (10, 11),
            "horn1_y": (12, 13),
            "beam_spot": (14, 16),
            "water_layer": (21, 22),
            "beam_shift_x": (24, 25),
            "beam_shift_y_plus": 26,
            "beam_shift_y_minus": 27,
            "beam_div": 32,
        }

        self.nominal_run = self._beam_pt[[15]]

        self.apply_systematic_selection()

    def is_excluded(self, run_id: int) -> bool:
        """Checks if run_id appears in the list of run_ids in the _run_id_map member variable.
        Returns True if it doesn't.
        """
        return not any(
            isinstance(item, int)
            and item == run_id
            or (isinstance(item, tuple) and len(item) == 2 and run_id in item)
            for item in self._run_id_map.values()
        )

    @cached_property
    def flux_shifts(self) -> pd.DataFrame:
        # All runs have a +/- 1sigma flux variant except for runs 30 and 32

        nom_vals = self.nominal_run.values

        beam_shifts = (self._beam_pt - nom_vals).drop(labels=[15], axis=1)

        beam_fractional_shifts = beam_shifts / nom_vals

        beam_fractional_shifts[beam_fractional_shifts.isna()] = 0

        if self.smoothing:
            beam_fractional_shifts = smooth_stat_fluctuations(
                beam_fractional_shifts, self.bin_edges
            )
            beam_shifts = beam_fractional_shifts * nom_vals

        flux_shifts_df = pd.concat(
            [beam_shifts, beam_fractional_shifts], keys=["absolute", "fractional"]
        )

        return flux_shifts_df

    def energy_to_bin_slice(self, elow, ehigh):
        index1 = np.argmax(self.bin_edges["numu"] >= elow)
        index2 = np.argmax(self.bin_edges["numu"] >= ehigh)
        return slice(index1, index2)

    def apply_systematic_selection(self):
        ids_to_drop = filter(self.is_excluded, self.flux_shifts.columns)
        self.flux_shifts.drop(ids_to_drop, axis=1, inplace=True)

        water_layer_indexer = (
            slice(None),
            slice(None),
            slice(None),
            self.energy_to_bin_slice(1.0, 20.0),
        ), (21, 22)
        self.flux_shifts.loc[water_layer_indexer] *= 0.0

        div_indexer = (
            slice(None),
            slice(None),
            slice(None),
            self.energy_to_bin_slice(0, 1.0),
        ), 32
        self.flux_shifts.loc[div_indexer] *= 0.0

    @cached_property
    def beam_systematic_shifts(self):
        flux_systs = {}
        shifts = self.flux_shifts.loc["absolute"]
        for key, run_id in self._run_id_map.items():
            if isinstance(run_id, tuple):
                id1, id2 = run_id
                flux_systs[key] = 0.5 * (shifts[id1] + shifts[id2])
                continue
            flux_systs[key] = shifts[run_id]

        df = pd.DataFrame(flux_systs, index=shifts.index)

        df_frac = df / self.nominal_run.values
        df_frac[df_frac.isna()] = 0

        beam_shifts = pd.concat(
            [df, df_frac],
            keys=["absolute", "fractional"],
            names=["scale"] + df.index.names,
        )

        beam_shifts.columns.name = "category"

        beam_shifts = (
            beam_shifts.stack("category")
            .reorder_levels(
                ["scale", "category", "horn_polarity", "neutrino_mode", "bin"]
            )
            .sort_index()
            .replace(-0.0, 0)
        )

        return beam_shifts

    @cached_property
    def covariance_matrices(self) -> pd.DataFrame:
        flux_shifts = self.beam_systematic_shifts.loc["absolute"]

        def _calc_cov(x):
            x = x.droplevel("category")
            outer = np.outer(x, x)
            return pd.DataFrame(outer, index=x.index, columns=x.index)

        covs = flux_shifts.groupby("category").apply(_calc_cov).replace(-0.0, 0.0)

        nom_mat = np.outer(self.nominal_run, self.nominal_run)

        grp_divide = lambda grp: np.divide(
            grp.droplevel(0), nom_mat, out=np.zeros_like(grp), where=nom_mat != 0
        )

        covs_frac = covs.groupby(level=0).apply(grp_divide)

        beam_covariance_matrices = pd.concat(
            [covs, covs_frac],
            keys=["absolute", "fractional"],
            names=["scale", "category", "horn_polarity", "neutrino_mode", "bin"],
        )

        # beam_covariance_matrices = beam_covariance_matrices.replace(
        #     -0.0, 0.0
        # ).sort_index()

        return beam_covariance_matrices

    @cached_property
    def correlation_matrices(self) -> pd.DataFrame:
        mat_groups = self.covariance_matrices.loc["absolute"].groupby(
            "category", group_keys=True
        )

        return mat_groups.apply(calculate_correlation_matrix)

    @cached_property
    def total_covariance_matrix(self) -> pd.DataFrame:
        covs_abs = self.covariance_matrices.loc["absolute"]
        beam_total_covariance_matrix_abs = (
            covs_abs.iloc[covs_abs.index.get_level_values("category") != "beam_power"]
            .groupby(level=["horn_polarity", "neutrino_mode", "bin"])
            .sum()
        )

        beam_total_covariance_matrix_frac = beam_total_covariance_matrix_abs / np.outer(
            self.nominal_run, self.nominal_run
        )

        total_covariance_matrix = pd.concat(
            [beam_total_covariance_matrix_frac, beam_total_covariance_matrix_abs],
            keys=["fractional", "absolute"],
            names=["scale", "horn_polarity", "neutrino_mode", "bin"],
        )

        return total_covariance_matrix

    @cached_property
    def total_correlation_matrix(self) -> pd.DataFrame | NDArray[Any]:
        corr = calculate_correlation_matrix(
            self.total_covariance_matrix.loc["absolute"]
        )
        return corr

    @cached_property
    def fractional_uncertainties(self) -> pd.DataFrame | pd.Series:
        levels = ["category", "horn_polarity", "neutrino_mode", "bin"]

        total_cov = self.total_covariance_matrix.loc["fractional"]
        total_cov["category"] = "total"
        total_cov = total_cov.set_index("category", append=True).reorder_levels(levels)

        covs = pd.concat([self.covariance_matrices.loc["fractional"], total_cov])

        cov_groups = covs.groupby(level=levels[0])

        # for some reason I get -0.0 for some values on the diagonal, so I'm wrapping in np.abs
        diag_sqrt = lambda x: pd.Series(np.sqrt(np.abs(np.diag(x))), index=covs.columns)

        return cov_groups.apply(diag_sqrt).stack(levels[1:]).sort_index()  # type: ignore



import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, NamedTuple, Optional

import cppyy
import numpy as np
import pandas as pd
import uproot
from ROOT import TFile  # type: ignore

TH1D = cppyy.gbl.TH1D


class HistInfo(NamedTuple):
    category: str
    neutrino: str
    universe: Optional[int] = None


def parse_th1_name(name: str) -> HistInfo:
    name = name[1:]
    split = name.split("_")

    match split:
        case ["nom", neutrino]:
            return HistInfo(category="nominal", neutrino=neutrino)
        case ["cv", neutrino]:
            return HistInfo(category="central_value", neutrino=neutrino)
        case ["nom", neutrino, category]:
            return HistInfo(category=category, neutrino=neutrino)
        case ["thin", *category, neutrino, universe]:
            return HistInfo(
                category="_".join(category), neutrino=neutrino, universe=int(universe)
            )
        case [category, neutrino, universe]:
            return HistInfo(
                category=category, neutrino=neutrino, universe=int(universe)
            )
        case _:
            raise ValueError(f"Cannot parse TH1 name: {name}")


def calculate_df(h: TH1D, horn: str, run_id: int, parsed: HistInfo) -> pd.DataFrame:
    nbins = len(h) - 2

    flux = np.zeros(nbins)
    stat_uncert = np.zeros(nbins)

    for i in range(1, nbins + 1):
        flux[i - 1] = h.GetBinContent(i)
        stat_uncert[i - 1] = h.GetBinError(i)

    df = pd.DataFrame(
        {
            "flux": flux,
            "stat_uncert": stat_uncert,
            "bin": range(1, nbins + 1),
            "category": parsed.category,
            "neutrino_mode": parsed.neutrino,
            "horn_polarity": horn,
            "run_id": run_id,
        }
    )

    if parsed.universe is not None:
        df["universe"] = parsed.universe

    return df


def normalize_flux_to_pot(
    input_file: str,
    horn: str,
    run_id: int,
    bin_edges: Optional[dict[str, np.ndarray]] = None,
    hist_name_filter: Optional[Callable[[str], bool]] = None,
) -> pd.DataFrame:
    """Normalizes flux histograms to POT and saves the data in a Pandas DataFrame.

    Args:
        input_file: Path to the input ROOT file containing the flux histograms and POT.
        horn: Horn polarity of the neutrino beam (either "FHC" or "RHC").
        run_id: ID number of the run for which the histograms were produced.
        bin_edges: Optional array of bin edges for rebinning the histograms.
        hist_name_filter: Optional function for filtering the histogram names.

    Returns:
        A Pandas DataFrame with columns for the flux, statistical uncertainty, bin number
    """

    logging.debug(f"Opening {input_file}...")

    with uproot.open(input_file) as f:  # type: ignore
        histkeys = f.keys(
            cycle=False,
            filter_classname="TH1D",
            filter_name=hist_name_filter,
        )

    with TFile(input_file) as f:
        pot = f.Get("hpot").GetMaximum()

        logging.debug(f"Normalizing to {pot} POT")

        hlist = []

        for key in histkeys:
            if "/" not in key:
                continue
            _, hist_name = key.rsplit("/", 1)

            parsed: HistInfo = parse_th1_name(hist_name)

            h = f.Get(key)

            if bin_edges is not None:
                logging.debug(f"Rebinning histogram {hist_name}")
                bins = bin_edges[parsed.neutrino]
                h = h.Rebin(len(bins) - 1, hist_name, bins)

            h.Scale(1.0 / pot)

            df = calculate_df(h, horn, run_id, parsed)
            logging.debug(df)

            logging.debug("Loading histogram into DataFrame")

            hlist.append(df)

    logging.debug(f"Closing {input_file}...")

    return pd.concat(hlist)



import logging
from concurrent.futures import ProcessPoolExecutor as Executor
from functools import partial

import pandas as pd
from rich.progress import Progress

from flux_tool.config import AnalysisConfig
from flux_tool.normalize_and_rebin_data import normalize_flux_to_pot


class Preprocessor:
    __slots__ = ("nominal_flux_df", "ppfx_correction_df")

    def __init__(self, cfg: AnalysisConfig) -> None:
        logging.info("\n=============== BEGINNING PREPROCESSING ===============")
        logging.info(
            "Reading input files, normalized to POT, and rebinning, if necessary"
        )
        jobs = list(cfg.itersamples())
        fn = partial(
            normalize_flux_to_pot,
            bin_edges=cfg.bin_edges,
            hist_name_filter=cfg.enabled_hist_filter,
        )
        results = []
        with Progress() as progress:
            task_id = progress.add_task("[cyan]Working...", total=len(jobs))
            with Executor() as executor:
                for job in jobs:
                    future = executor.submit(fn, *job)
                    future.add_done_callback(lambda _: progress.advance(task_id))
                    results.append(future.result())

        df = pd.concat(results)
        self.nominal_flux_df = df.loc[df["universe"].isna()].drop("universe", axis=1)
        is_unis = df["universe"].notna()
        self.ppfx_correction_df = df.loc[is_unis]



from flux_tool.vis_scripts.covariance import (plot_beam_correlation_matrices,
                                              plot_beam_covariance_matrices,
                                              plot_hadron_correlation_matrices,
                                              plot_hadron_covariance_matrices)
from flux_tool.vis_scripts.flux_prediction import (
    plot_flux_prediction, plot_flux_uncorrected_logarithmic)
from flux_tool.vis_scripts.fractional_uncertainties import (
    plot_beam_fractional_uncertainties, plot_beam_systematic_shifts,
    plot_hadron_fractional_uncertainties,
    plot_hadron_fractional_uncertainties_mesinc_breakout,
    plot_hadron_fractional_uncertainties_nua_breakout,
    plot_uncertainties
    )
from flux_tool.vis_scripts.parent_spectra import plot_parents
from flux_tool.vis_scripts.pca_plots import (
    pca_mesinc_overlay, plot_hadron_systs_and_pca_variances,
    plot_pca_systematic_shifts, plot_top_principal_components, scree_plot)
from flux_tool.vis_scripts.ppfx_universes import plot_ppfx_universes



import logging
import sys
import tomllib
from datetime import date
from pathlib import Path
from typing import Generator, Self

import numpy as np


class AnalysisConfig:
    """
    Configuration class for analysis parameters and paths.

    This class encapsulates the configuration settings and paths required
    for performing analysis. It allows setting up binning, paths for sources,
    results, plots, and handling ignored histogram names. Additionally, it
    provides methods for parsing filenames, iterating through samples, and
    loading configuration from strings or files.

    Parameters:
    project_config (dict): A dictionary containing the project configuration
        settings, including plotting options, binning, paths, and more.

    Attributes:
    bin_edges (dict): A dictionary containing the bin edges for different
        neutrino types.
    neutrinos (list): A list of neutrino types: ["nue", "nuebar", "numu", "numubar"].
    enabled_histogram_names: List of ignored histogram names based on specification under [PPFX] in the config.toml.
    samples (dict): A dictionary containing paths to samples for "fhc" and "rhc" horn operating modes.
    output_file_name (str): The name of the output file.
    plot_opts (dict): Plotting options, such as x-axis limits.
    plots_path (Path): Path to the directory where plots will be saved.
    products_file (str): Path to the products file.
    results_path (Path): Path to the directory where analysis results will be saved.
    inputs_path (Path): Path to the directory containing analysis sources.

    Methods:
    verify_paths(): Verifies the existence of necessary paths and creates them if missing.
    enabled_hist_filter(hist_name: str) -> bool: Checks if a histogram name should be ignored.
    itersamples(): Iterates through source files, yielding file, horn current, and run ID.
    from_str(config_str: str) -> AnalysisConfig: Creates an instance from a configuration string.
    from_file(config_file: str) -> AnalysisConfig: Creates an instance from a configuration file.
    """

    __slots__ = (
        "bin_edges",
        "enabled_histogram_names",
        "neutrinos",
        "output_file_name",
        "plot_opts",
        "plots_path",
        "products_file",
        "results_path",
        "samples",
        "inputs_path",
    )

    def __init__(self, project_config: dict) -> None:
        plotting = project_config["Plotting"]

        self.plot_opts = {
            "draw_label": plotting["draw_label"],
            "experiment": plotting["experiment"],
            "stage": plotting["stage"],
            "xlim": plotting.get("neutrino_energy_range", (0.0, 20.0)),
            "flux_prediction_bullets": plotting["flux_prediction_bullets"],
            "enabled": plotting["enabled"],
        }

        self.neutrinos: list[str] = ["nue", "nuebar", "numu", "numubar"]

        def_binning = {nu: np.linspace(0, 20, num=201) for nu in self.neutrinos}

        binning = project_config.get("Binning", def_binning)

        self.bin_edges = {}

        for nu, bins in binning.items():
            if isinstance(bins, int):
                self.bin_edges[nu] = np.linspace(0, 20, num=bins + 1)
            elif isinstance(bins, list) and isinstance(bins[0], float):
                self.bin_edges[nu] = np.asarray(bins)
            elif (
                isinstance(bins, list)
                and isinstance(bins[0], list)
                and len(bins[0]) == 3
            ):
                tmp = np.array([])
                for start, stop, step in bins:
                    tmp = np.append(tmp, np.arange(start, stop, step))
                self.bin_edges[nu] = tmp
            else:
                logging.error(
                    f"Invalid binning for {nu}. Falling back to default binning."
                )
                self.bin_edges[nu] = def_binning[nu]

        self.inputs_path = (
            Path(project_config["Inputs"]["directory"]).expanduser().resolve()
        )

        self.results_path = Path(
            project_config.get(
                "results",
                self.inputs_path.parent,
            )
        )

        self.plots_path = Path(
            project_config.get(
                "plots",
                self.inputs_path.parent / "plots/",
            )
        )

        output_file = project_config["output_file_name"]

        self.products_file = f"{self.results_path}/{date.today()}_{output_file}"

        self.enabled_histogram_names = []
        for k, v in project_config["PPFX"]["enabled"].items():
            if v:
                continue
            if k == "thintarget":
                self.enabled_histogram_names += ["hthin_nue", "hthin_numu"]
                continue
            if k == "mippnumi":
                self.enabled_histogram_names.append("mipp")
                continue
            self.enabled_histogram_names.append(k)

        inputs = {k: v for k, v in project_config["Inputs"].items() if k != "directory"}

        self.samples = {}
        for horn, input in inputs.items():
            self.samples[horn] = {}
            for name, sample in input.items():
                self.samples[horn][name] = self.inputs_path / Path(sample)

    def verify_paths(self) -> None:
        for path in [self.inputs_path, self.results_path, self.plots_path]:
            if not path.exists():
                opt = input(f"{path} does not exist. Create it? (y/n) ").lower()
                if opt == "y":
                    path.mkdir()
                else:
                    print("Directory not created. Exiting...")
                    sys.exit()

        if not any(self.inputs_path.iterdir()):
            msg = (
                f'No files found in input directory: "{self.inputs_path}"'
                "\nExiting..."
            )
            raise FileNotFoundError(msg)

    def enabled_hist_filter(self, hist_name: str) -> bool:
        return not any(
            x.lower() in hist_name.lower() for x in self.enabled_histogram_names
        )

    def itersamples(self) -> Generator[tuple[str, str, int], None, None]:
        for horn, samples in self.samples.items():
            for name, sample in samples.items():
                yield str(sample), horn, name

    @classmethod
    def from_str(cls, config_str: str) -> Self:
        config = tomllib.loads(config_str)
        return cls(config)

    @classmethod
    def from_file(cls, config_file: str) -> Self:
        with open(config_file, "rb") as file:
            config = tomllib.load(file)
        logging.info(f"Read configuration from {config_file}")
        return cls(config)

    @staticmethod
    def dump_default_config() -> str:
        return """\
# flux_tool configuration file

output_file_name = "out.root"

[Inputs]
directory = "/path/to/directory/containing/input/histograms"
fhc.nominal = "input_fhc_nominal.root"
rhc.nominal = "input_rhc_nominal.root"
fhc.horn_current_up = "input_fhc_horn_current_up.root"
fhc.horn_current_down = "input_fhc_horn_current_down.root"

[Binning]
# Histogram bin edges for each neutrino flavor.
# Accepts:
#    1. an integer number of bins (between 0 and 20 GeV)
#    2. An array of bin edges (NOTE: they can be variable bin widths, but must be monotonically increasing)
#    3. An array of arrays where the inner arrays are [start, stop, step] for fixed bin widths.
#    4. If unspecified, then fixed bin widths of 100 MeV is applied along the [0, 20] GeV interval.
nue = 200

nuebar = [
  0.0,
  0.2,
  0.4,
  0.6,
  0.8,
  1.0,
  1.5,
  2.0,
  2.5,
  3.0,
  3.5,
  4.0,
  6.0,
  8.0,
  12.0,
]

numu = [[0.0, 6.0, 0.1], [6.0, 20.0, 0.5]]

numubar = [[0.0, 6.0, 0.1], [6.0, 20.0, 0.5]]

  [PPFX]
# enable/disable specific PPFX reweight categories from
# appearing in the fractional uncertainty directory
# true = included, false = excluded
[PPFX.enabled]
total = true
attenuation = true
mesinc = true
mesinc_parent_K0 = true
mesinc_parent_Km = true
mesinc_parent_Kp = true
mesinc_parent_pim = true
mesinc_parent_pip = true
mesinc_daughter_K0 = true
mesinc_daughter_Km = true
mesinc_daughter_Kp = true
mesinc_daughter_pim = true
mesinc_daughter_pip = true
mippnumi = false
nua = true
pCfwd = false
pCk = true
pCpi = true
pCnu = true
pCQEL = false
others = true
thintarget = false

[Plotting]
draw_label = true                   # whether or not to draw the experiment label, e.g., ICARUS Preliminary
experiment = "ICARUS"
stage = "Preliminary"
neutrino_energy_range = [0.0, 6.0]  # horizontal axis limits in [GeV]
flux_prediction_bullets = false     # whether or not to draw bullets or lines with error band for flux prediction

[Plotting.enabled]
# Enable/disable specific plots from the visualization output
uncorrected_flux = true
flux_prediction = true
flux_prediction_parent_spectra = true
flux_prediction_parent_spectra_stacked = true
ppfx_universes = true
hadron_uncertainties = true
hadron_uncertainties_meson = true
hadron_uncertainties_meson_only = true
pca_scree_plot = true
pca_mesinc_overlay = true
pca_top_components = true
pca_variances = true
pca_components = true
hadron_covariance_matrices = "total"
hadron_correlation_matrices = true
beam_uncertainties = true
beam_covariance_matrices = true
beam_correlation_matrices = true
beam_systematic_shifts = true
"""



from pathlib import Path
from typing import Optional

import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np

from flux_tool.vis_scripts.helper import create_ylabel_with_scale, save_figure
from flux_tool.vis_scripts.spectra_reader import SpectraReader
from flux_tool.vis_scripts.style import neutrino_labels, xlabel_enu


def plot_ppfx_universes(
    reader: SpectraReader,
    output_dir: Optional[Path] = None,
    xlim: tuple[float, float] = (0.0, 20.0),
):
    nominal_flux = reader.nominal_spectra
    universes = reader.universes
    ppfx_correction = reader.ppfx_correction

    pot = reader.pot

    for horn, nu in reader.horns_and_nus:
        unis = [universes[f"{horn}/{nu}_total/htotal_{nu}_{x}"] for x in range(100)]
        nom = nominal_flux[f"{horn}/nom/hnom_{nu}"]
        correction = ppfx_correction[f"htotal_{horn}_{nu}"].to_pyroot()

        _, bins = ppfx_correction[f"htotal_{horn}_{nu}"].to_numpy()

        max_flux = correction.GetMaximum()

        power = -1 * np.round(np.log10(max_flux))

        scale_factor = 10**power

        correction.Scale(scale_factor)

        def _scale(h):
            h = h.to_pyroot()
            h = h.Rebin(len(bins) - 1, "", bins)
            h.Scale(scale_factor / pot[horn])
            return h

        H = list(map(_scale, unis))

        nom_scaled = _scale(nom)

        fig, ax = plt.subplots()

        ax.set_box_aspect(1)

        ax.plot([], [], " ", label="PPFX Input")

        # nu_label = neutrino_labels[nu]

        hep.histplot(
            H=nom_scaled,
            label="Uncorrected Flux",
            # bins=bins,
            color="k",
            yerr=False,
            edges=False,
            binwnorm=1,
            ls="--",
            zorder=10,
            ax=ax,
        )

        ax.plot([], [], " ", label="PPFX Output")

        hep.histplot(
            H=H[0],
            label="Flux Universes",
            color="C4",
            lw=1,
            yerr=False,
            edges=False,
            binwnorm=1,
            ax=ax,
        )
        hep.histplot(
            H=H[1:],
            color="C4",
            lw=1,
            yerr=False,
            edges=False,
            binwnorm=1,
            ax=ax,
        )

        hep.histplot(
            H=correction,
            label=r"Mean Flux ($\pm \mathrm{\sigma}$)",
            color="C0",
            histtype="errorbar",
            xerr=False,
            capsize=4,
            elinewidth=4,
            marker="o",
            binwnorm=1,
            ax=ax,
            zorder=15,
        )

        ax.set_xlim(xlim)
        ax.set_xlabel(xlabel_enu)
        ax.set_ylabel(create_ylabel_with_scale(int(power)))

        handles, labels = ax.get_legend_handles_labels()

        leg = ax.legend(
            handles,
            labels,
            loc="best",
        )

        hep.label.exp_label(
            exp="NuMI",
            llabel="Simulation",
            rlabel="",
        )

        ax.annotate(
            f"{horn.upper()} {neutrino_labels[nu]}",
            xy=(1.00, 1.00),
            xycoords="axes fraction",
            ha="right",
            va="bottom",
            fontweight="bold",
        )

        for item, label in zip(leg.legend_handles, leg.texts):
            if label._text in ["PPFX Input", "PPFX Output"]:
                width = item.get_window_extent(fig.canvas.get_renderer()).width  # type: ignore
                label.set_ha("left")
                label.set_position((-2 * width, 0))

        if output_dir is not None:
            prefix = f"{horn}_{nu}"
            file_stem = f"{prefix}_ppfx_universes"
            tex_caption = ""
            tex_label = file_stem
            save_figure(fig, file_stem, output_dir, tex_caption, tex_label)  # type: ignore



import numpy as np
import pandas as pd


def extract_statistical_uncertainties(
    nominal_dataframe: pd.DataFrame, flux_weights: pd.DataFrame, normalized=False
) -> pd.Series:
    index = ("category", "horn_polarity", "neutrino_mode", "bin")
    pivot_table = pd.pivot_table(
        nominal_dataframe, index=index, values=["flux", "stat_uncert"]
    )

    pt = pivot_table.loc["nominal"].mul(flux_weights, axis=0)

    stats = pt["stat_uncert"]

    if normalized:
        return stats / pt["flux"]

    return stats


def flux_uncertainty(cov: pd.DataFrame, total_flux: float) -> float:
    return np.sqrt(cov.sum().sum() / total_flux**2)


def ratio_uncertainty(
    cov: pd.DataFrame, total_nue_flux: float, total_numu_flux: float
) -> float:
    reordered_mat = cov.stack("neutrino_mode").swaplevel(i=1, j=2).sort_index()
    if reordered_mat is None:
        raise ValueError("cov is not of the expected format.")

    mat_groups = reordered_mat.groupby(level=(0, 1))

    total = 0.0
    for (nu1, nu2), mat in mat_groups:  # type: ignore
        cov_sum = mat.sum().sum()
        if nu1 in ["nue", "nuebar"] and nu2 in ["nue", "nuebar"]:
            div = total_nue_flux**-2
            coeff = 1
        elif nu1 in ["numu", "numubar"] and nu2 in ["numu", "numubar"]:
            div = total_numu_flux**-2
            coeff = 1
        else:
            div = 1 / (total_nue_flux * total_numu_flux)
            coeff = -1
        total += coeff * cov_sum * div

    return np.sqrt(total)
